# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Command line client.
"""


from collections import defaultdict, namedtuple
from copy import deepcopy
from datetime import datetime, timedelta
import hashlib
import os
import re
import sys
import shutil
import time
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union

from codechecker_api.codeCheckerDBAccess_v6 import constants, ttypes
from codechecker_api_shared.ttypes import RequestFailed

from codechecker_report_converter import twodim
from codechecker_report_converter.report import File, Report, report_file, \
    reports as reports_helper
from codechecker_report_converter.report.output import baseline, codeclimate, \
    gerrit, json as report_to_json, plaintext
from codechecker_report_converter.report.output.html import \
    html as report_to_html
from codechecker_report_converter.report.statistics import Statistics
from codechecker_report_converter.util import dump_json_output

from codechecker_common import logger
from codechecker_common.checker_labels import CheckerLabels
from codechecker_common.review_status_handler import ReviewStatusHandler
from codechecker_common.util import load_json

from codechecker_web.shared import convert, webserver_context

from codechecker_client import report_type_converter
from .client import login_user, setup_client
from .cmd_line import CmdLineOutputEncoder
from .product import split_server_url

from . import suppress_file_handler

# Needs to be set in the handler functions.
LOG = None


BugPathLengthRange = namedtuple('BugPathLengthRange', ['min', 'max'])


def init_logger(level, stream=None, logger_name='system'):
    logger.setup_logger(level, stream)
    global LOG
    LOG = logger.get_logger(logger_name)


def filter_local_file_remote_run(
    run_args: List[str]
) -> Tuple[List[str], List[str], List[str]]:
    """
    Filter out arguments which are local directory, baseline files or remote
    run names.
    """
    local_dirs = []
    baseline_files = []
    run_names = []

    for r in run_args:
        if os.path.isdir(r):
            local_dirs.append(os.path.abspath(r))
        elif os.path.isfile(r) and baseline.check(r):
            baseline_files.append(os.path.abspath(r))
        else:
            run_names.append(r)

    return local_dirs, baseline_files, run_names


def run_sort_type_str(value: ttypes.RunSortType) -> Optional[str]:
    """ Converts the given run sort type to string. """
    if value == ttypes.RunSortType.NAME:
        return 'name'
    elif value == ttypes.RunSortType.UNRESOLVED_REPORTS:
        return 'unresolved_reports'
    elif value == ttypes.RunSortType.DATE:
        return 'date'
    elif value == ttypes.RunSortType.DURATION:
        return 'duration'
    elif value == ttypes.RunSortType.CC_VERSION:
        return 'codechecker_version'


def run_sort_type_enum(value: str) -> Optional[ttypes.RunSortType]:
    """ Returns the given run sort type Thrift enum value. """
    if value == 'name':
        return ttypes.RunSortType.NAME
    elif value == 'unresolved_reports':
        return ttypes.RunSortType.UNRESOLVED_REPORTS
    elif value == 'date':
        return ttypes.RunSortType.DATE
    elif value == 'duration':
        return ttypes.RunSortType.DURATION
    elif value == 'codechecker_version':
        return ttypes.RunSortType.CC_VERSION


def get_diff_type(args) -> Union[ttypes.DiffType, None]:
    """
    Returns Thrift DiffType value by processing the arguments.
    """
    if 'new' in args:
        return ttypes.DiffType.NEW

    if 'unresolved' in args:
        return ttypes.DiffType.UNRESOLVED

    if 'resolved' in args:
        return ttypes.DiffType.RESOLVED

    return None


def get_run_tag(client, run_ids: List[int], tag_name: str):
    """Return run tag information for the given tag name in the given runs."""
    run_history_filter = ttypes.RunHistoryFilter()
    run_history_filter.tagNames = [tag_name]
    run_histories = client.getRunHistory(run_ids, None, None,
                                         run_history_filter)

    return run_histories[0] if run_histories else None


def process_run_args(client, run_args_with_tag: Iterable[str]):
    """Process the argument and returns run ids and run tag ids.

    The elemnts inside the given run_args_with_tag list has the following
    format: `<run_name>:<run_tag>`.
    If the run name is containing a literal ':', it should be escaped with a
    backslash escape character before the `:`.
    """
    # (The documentation comment above means \:, but Python would want to
    # understand that as an escape sequence immediately.)

    run_ids = []
    run_names = []
    run_tags = []

    for run_arg_with_tag in run_args_with_tag:
        run_with_tag = list(map(lambda n: n.replace(r"\:", ":"),
                                re.split(r"(?<!\\):", run_arg_with_tag)))
        run_name = run_with_tag[0]
        run_filter = ttypes.RunFilter(names=[run_name])

        runs = get_run_data(client, run_filter)
        if not runs:
            LOG.warning("No run names match the given pattern: %s",
                        run_arg_with_tag)
            sys.exit(1)

        r_ids = [run.runId for run in runs]
        run_ids.extend(r_ids)

        r_names = [run.name for run in runs]
        run_names.extend(r_names)

        # Set base run tag if it is available.
        run_tag_name = run_with_tag[1] if len(run_with_tag) > 1 else None
        if run_tag_name:
            tag = get_run_tag(client, r_ids, run_tag_name)
            if tag:
                run_tags.append(tag.id)

    return run_ids, run_names, run_tags


def get_suppressed_reports(reports: List[Report],
                           review_st: List[ttypes.ReviewStatus]) -> List[str]:
    """Returns a list of suppressed report hashes."""
    statuses_str = [ttypes.ReviewStatus._VALUES_TO_NAMES[x].lower()
                    for x in review_st]

    return [
        report.report_hash for report in reports
        if report.review_status.status != 'unreviewed' and
        report.review_status not in statuses_str]


def get_report_dir_results(
    report_dirs: List[str],
    report_filter: ttypes.ReportFilter,
    checker_labels: CheckerLabels
) -> List[Report]:
    """Get reports from the given report directories.

    Absolute paths are expected to the given report directories.
    """
    all_reports = []

    processed_path_hashes = set()
    for report_dir, file_paths in \
            report_file.analyzer_result_files(report_dirs):

        review_status_handler = ReviewStatusHandler()
        review_status_cfg = os.path.join(report_dir, 'review_status.yaml')
        if os.path.isfile(review_status_cfg):
            try:
                review_status_handler.set_review_status_config(
                    review_status_cfg)
            except ValueError as err:
                LOG.error(err)
                sys.exit(1)

        for file_path in file_paths:
            # Get reports.
            reports = report_file.get_reports(file_path, checker_labels)

            try:
                for report in reports:
                    report.review_status = \
                        review_status_handler.get_review_status(report)
            except ValueError as err:
                LOG.error(err)

            # Skip duplicated reports.
            reports = reports_helper.skip(reports, processed_path_hashes)

            # Skip reports based on filter arguments.
            reports = [report for report in reports
                       if not skip_report_dir_result(
                           report, report_filter, checker_labels)]

            all_reports.extend(reports)

    return all_reports


def skip_report_dir_result(
    report: Report,
    report_filter: ttypes.ReportFilter,
    checker_labels: CheckerLabels
) -> bool:
    """Returns True if the report should be skipped from the results.

    Skipping is done based on the given filter set.
    """

    if report_filter.severity:
        severity_name = checker_labels.severity(report.checker_name)
        filter_severities_str = [ttypes.Severity._VALUES_TO_NAMES[x]
                                 for x in report_filter.severity]
        if severity_name.lower() not in \
                list(map(str.lower, filter_severities_str)):
            return True

    if report_filter.checkerName:
        checker_name = report.checker_name
        if not any([re.match(r'^' + c.replace("*", ".*") + '$',
                             checker_name, re.IGNORECASE)
                    for c in report_filter.checkerName]):
            return True

    if report_filter.filepath:
        if not any([re.match(r'^' + f.replace("*", ".*") + '$',
                             report.file.path, re.IGNORECASE)
                    for f in report_filter.filepath]):
            return True

    if report_filter.checkerMsg:
        checker_msg = report.message
        if not any([re.match(r'^' + c.replace("*", ".*") + '$',
                             checker_msg, re.IGNORECASE)
                    for c in report_filter.checkerMsg]):
            return True

    return False


def get_diff_base_results(
    client,
    report_filter: ttypes.ReportFilter,
    baseids,
    base_hashes,
    suppressed_hashes,
    get_details: bool = True
):
    """Get the run results from the server."""
    report_filter.reportHash = base_hashes + suppressed_hashes
    report_filter.reviewStatus = [
        ttypes.ReviewStatus.UNREVIEWED,
        ttypes.ReviewStatus.CONFIRMED]

    sort_mode = [(ttypes.SortMode(
        ttypes.SortType.FILENAME,
        ttypes.Order.ASC))]

    limit = constants.MAX_QUERY_SIZE
    offset = 0

    base_results = []
    while True:
        results = client.getRunResults(
            baseids, limit, offset, sort_mode, report_filter, None,
            get_details)

        base_results.extend(results)
        offset += limit

        if len(results) < limit:
            break

    return base_results


def str_to_timestamp(date_str):
    """Return timestamp parsed from the given string parameter."""
    dateformat = '%Y-%m-%d %H:%M:%S.%f'
    date_time = date_str if isinstance(date_str, datetime) else \
        datetime.strptime(date_str, dateformat)

    return time.mktime(date_time.timetuple())


def check_run_names(client, check_names):
    """Check if the given names are valid runs on the server.

    If any of the names is not found then the script finishes.
    Otherwise a dictionary returns which maps run names to runs.
    The dictionary contains only the runs in check_names or
    all runs if check_names is empty or None.
    """
    run_filter = ttypes.RunFilter()
    run_filter.names = check_names
    run_filter.exactMatch = check_names is not None

    run_info = {run.name: run for run in get_run_data(client, run_filter)}

    if not check_names:
        return run_info

    missing_names = [name for name in check_names if not run_info.get(name)]
    if missing_names:
        LOG.warning("The following runs were not found in the database: %s.",
                    ', '.join(missing_names))
        sys.exit(1)

    return run_info


def check_existing_source_components(
    client,
    source_components: List[str]
):
    missing_source_components = []

    for source_component in source_components:
        if not client.getSourceComponents([source_component]):
            missing_source_components.append(source_component)

    if missing_source_components:
        LOG.error(
            "The following source components don't exist: %s",
            ', '.join(missing_source_components))
        sys.exit(1)


def check_deprecated_arg_usage(args):
    if 'detected_at' in args:
        LOG.warning('"--detected-at" option has been deprecated. Use '
                    '--detected-before/--detected-after options to filter the '
                    'results by detection date. For more information see the '
                    'help.')

    if 'fixed_at' in args:
        LOG.warning('"--fixed-at" option has been deprecated. Use '
                    '--fixed-before/--fixed-after options to filter the '
                    'results by fix date. For more information see the '
                    'help.')


def get_run_data(client, run_filter, sort_mode=None,
                 limit=constants.MAX_QUERY_SIZE):
    """Get all runs based on the given run filter."""

    all_runs = []

    offset = 0
    while True:
        runs = runs = client.getRunData(run_filter, limit, offset, sort_mode)
        all_runs.extend(runs)
        offset += limit

        if len(runs) < limit:
            break

    return all_runs


def get_run_results(client,
                    run_ids,
                    limit,
                    offset,
                    sort_type,
                    report_filter,
                    compare_data,
                    query_report_details):
    """Get all the results with multiple api request.

    In each api request get the limit ammount of reports.
    Collect and return all the reports based on the filters.
    """

    all_results = []
    while True:
        results = client.getRunResults(run_ids,
                                       limit,
                                       offset,
                                       sort_type,
                                       report_filter,
                                       compare_data,
                                       query_report_details)
        all_results.extend(results)
        offset += limit
        if len(results) < limit:
            break

    return all_results


def validate_filter_values(converted_values: List[int], valid_values,
                           value_type):
    """
    Check if the value provided by the user is a valid value.
    """
    # Can be None if the user provided no value to this option -- that is
    # valid.
    if not converted_values:
        return True

    non_valid_values = [status for status in converted_values
                        if valid_values.get(status, None) is None]

    if non_valid_values:
        invalid_values = ','.join([x.lower() for x in non_valid_values])
        LOG.error('Invalid %s value(s): %s', value_type, invalid_values)

        valid_values = ','.join([x.lower() for x in valid_values.keys()])
        LOG.error('Valid values are: %s', valid_values)

        return False

    return True


def parse_report_filter(client, args):
    """
    Parse and check attributes of the given report filter based on
    the arguments which is provided in the command line.
    Also, check if filter values are valid values.
    """
    report_filter = parse_report_filter_offline(args)

    if 'tag' in args:
        run_history_filter = ttypes.RunHistoryFilter(tagNames=args.tag)
        run_histories = client.getRunHistory(None, None, None,
                                             run_history_filter)
        if run_histories:
            report_filter.runTag = [t.id for t in run_histories]

    return report_filter


def parse_report_filter_offline(args):
    """
    Same as parse_report_filter, but will not make calls to the server.
    As of writing this comment, this means that the 'tag' argument will be
    ignored (as that would require a getRunHistory API call).
    """
    report_filter = ttypes.ReportFilter()

    if 'severity' in args:
        report_filter.severity = [
            ttypes.Severity._NAMES_TO_VALUES[x.upper()] for x in args.severity]

    if 'detection_status' in args:
        report_filter.detectionStatus = [
            ttypes.DetectionStatus._NAMES_TO_VALUES[x.upper()] for x in
            args.detection_status]

    if 'review_status' in args:
        report_filter.reviewStatus = [
            ttypes.ReviewStatus._NAMES_TO_VALUES[x.upper()] for x in
            args.review_status]

    if 'checker_name' in args:
        report_filter.checkerName = args.checker_name

    if 'file_path' in args:
        report_filter.filepath = args.file_path

    if 'bug_path_length' in args:
        path_lengths = args.bug_path_length.split(':')

        min_bug_path_length = int(path_lengths[0]) if \
            path_lengths and path_lengths[0].isdigit() else None

        max_bug_path_length = int(path_lengths[1]) if \
            len(path_lengths) > 1 and path_lengths[1].isdigit() else None

        report_filter.bugPathLength = \
            BugPathLengthRange(min=min_bug_path_length,
                               max=max_bug_path_length)

    values_to_check = [
        (report_filter.severity, ttypes.Severity._VALUES_TO_NAMES, 'severity'),
        (report_filter.detectionStatus,
         ttypes.DetectionStatus._VALUES_TO_NAMES,
         'detection status'),
        (report_filter.reviewStatus, ttypes.ReviewStatus._VALUES_TO_NAMES,
         'review status')]

    if not all(valid for valid in
               [validate_filter_values(*x) for x in values_to_check]):
        sys.exit(1)

    report_filter.isUnique = args.uniqueing == 'on'

    if 'checker_msg' in args:
        report_filter.checkerMsg = args.checker_msg

    if 'analyzer_name' in args:
        report_filter.analyzerNames = args.analyzer_name

    if 'component' in args:
        report_filter.componentNames = args.component

    if 'report_hash' in args:
        report_filter.reportHash = args.report_hash

    if 'open_reports_date' in args:
        report_filter.openReportsDate = \
            int(str_to_timestamp(args.open_reports_date))

    detected_at = None
    fixed_at = None

    if 'detected_at' in args:
        report_filter.firstDetectionDate = \
            int(str_to_timestamp(args.detected_at))

    if 'fixed_at' in args:
        report_filter.fixDate = int(str_to_timestamp(args.fixed_at))

    if 'detected_before' in args or 'detected_after' in args:
        detected_at = ttypes.DateInterval()

        if 'detected_before' in args:
            detected_at.before = int(str_to_timestamp(args.detected_before))

        if 'detected_after' in args:
            detected_at.after = int(str_to_timestamp(args.detected_after))

    if 'fixed_before' in args or 'fixed_after' in args:
        fixed_at = ttypes.DateInterval()

        if 'fixed_before' in args:
            fixed_at.before = int(str_to_timestamp(args.fixed_before))

        if 'fixed_after' in args:
            fixed_at.after = int(str_to_timestamp(args.fixed_after))

    if detected_at or fixed_at:
        report_filter.date = ttypes.ReportDate(detected=detected_at,
                                               fixed=fixed_at)

    report_filter.fileMatchesAnyPoint = args.anywhere_on_report_path
    report_filter.componentMatchesAnyPoint = args.anywhere_on_report_path

    if args.anywhere_on_report_path and \
            'file_path' not in args and 'component' not in args:
        LOG.warning(
            'The flag --anywhere-on-report-path is meaningful only if --file '
            'or --component is used.')

    return report_filter


def process_run_filter_conditions(args):
    """
    This function fills some attributes of the given run filter based on
    the arguments which is provided in the command line.
    """
    run_filter = ttypes.RunFilter()

    if 'names' in args:
        run_filter.names = args.names
        run_filter.exactMatch = False
    elif 'all_after_run' in args:
        run_filter.afterRun = args.all_after_run
    elif 'all_before_run' in args:
        run_filter.beforeRun = args.all_before_run
    elif 'all_after_time' in args:
        run_filter.afterTime = int(str_to_timestamp(args.all_after_time))
    elif 'all_before_time' in args:
        run_filter.beforeTime = int(str_to_timestamp(args.all_before_time))

    return run_filter


# ---------------------------------------------------------------------------
# Argument handlers for the 'CodeChecker cmd' subcommands.
# ---------------------------------------------------------------------------


def handle_list_runs(args):
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args and args.output_format != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    client = setup_client(args.product_url)

    run_filter = process_run_filter_conditions(args)

    sort_type = run_sort_type_enum(args.sort_type)
    sort_order = ttypes.Order._NAMES_TO_VALUES[args.sort_order.upper()]
    sort_mode = ttypes.RunSortMode(sort_type, sort_order)

    runs = get_run_data(client, run_filter, sort_mode)

    if args.output_format == 'json':
        # This json is different from the json format printed by the
        # parse command. This json converts the ReportData type report
        # to a json format.
        results = []
        for run in runs:
            if 'details' in args and args.details:
                run.analyzerStatistics = \
                    client.getAnalysisStatistics(run.runId, None)
            results.append({run.name: run})
        print(CmdLineOutputEncoder().encode(results))

    else:  # plaintext, csv
        header = ['Name', 'Number of unresolved reports',
                  'Analyzer statistics', 'Storage date', 'Version tag',
                  'Duration', 'Description', 'CodeChecker version']
        rows = []
        for run in runs:
            duration = str(timedelta(seconds=run.duration)) \
                if run.duration > -1 else 'Not finished'

            analyzer_statistics = []
            for analyzer in run.analyzerStatistics:
                stat = run.analyzerStatistics[analyzer]
                num_of_all_files = stat.successful + stat.failed
                analyzer_statistics.append(analyzer + ' (' +
                                           str(num_of_all_files) + '/' +
                                           str(stat.successful) + ')')

            codechecker_version = run.codeCheckerVersion \
                if run.codeCheckerVersion else ''
            description = run.description if run.description else ''

            rows.append((run.name,
                         str(run.resultCount),
                         ', '.join(analyzer_statistics),
                         run.runDate,
                         run.versionTag if run.versionTag else '',
                         duration,
                         description,
                         codechecker_version))

        print(twodim.to_str(args.output_format, header, rows))


def handle_list_results(args):
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args and args.output_format != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    check_deprecated_arg_usage(args)

    client = setup_client(args.product_url)

    if 'component' in args:
        check_existing_source_components(client, args.component)

    run_filter = ttypes.RunFilter(names=args.names)

    run_data = get_run_data(client, run_filter)
    run_ids = [run.runId for run in run_data]
    if not run_ids:
        LOG.warning("No runs were found!")
        sys.exit(1)

    report_filter = parse_report_filter(client, args)

    query_report_details = args.details and args.output_format == 'json' \
        if 'details' in args else None

    all_results = get_run_results(client,
                                  run_ids,
                                  constants.MAX_QUERY_SIZE,
                                  0,
                                  None,
                                  report_filter,
                                  None,
                                  query_report_details)

    if args.output_format == 'json':
        if 'details' in args:
            run_id2name = {run.runId: run.name for run in run_data}
            for report in all_results:
                report.runName = run_id2name[report.runId]

        print(CmdLineOutputEncoder().encode(all_results))
    else:
        header = ['File', 'Checker', 'Severity', 'Message', 'Bug path length',
                  'Analyzer name', 'Review status', 'Detection status']

        rows = []
        max_msg_len = 50
        for res in all_results:
            bug_line = res.line
            checked_file = res.checkedFile
            if bug_line is not None:
                checked_file += ' @ ' + str(bug_line)

            sev = ttypes.Severity._VALUES_TO_NAMES[res.severity]
            rw_status = \
                ttypes.ReviewStatus._VALUES_TO_NAMES[res.reviewData.status]

            dt_status = 'N/A'

            status = res.detectionStatus
            if status is not None:
                dt_status = ttypes.DetectionStatus._VALUES_TO_NAMES[status]

            # Remove whitespace characters from the checker message.
            msg = re.sub(r'\s+', ' ', res.checkerMsg)

            # Avoid too long cell content.
            if len(msg) > max_msg_len:
                msg = msg[:max_msg_len] + '...'

            rows.append((checked_file, res.checkerId, sev, msg,
                         res.bugPathLength, res.analyzerName, rw_status,
                         dt_status))

        print(twodim.to_str(args.output_format, header, rows))


def get_source_line_contents(
    client,
    reports: List[ttypes.ReportData]
) -> Dict[int, Dict[int, str]]:
    """
    Get source line contents from the server for the given report data.
    """
    source_lines = defaultdict(set)
    for report_data in reports:
        source_lines[report_data.fileId].add(report_data.line)

    lines_in_files_requested = []
    for file_id in source_lines:
        lines_in_files_requested.append(
            ttypes.LinesInFilesRequested(fileId=file_id,
                                         lines=source_lines[file_id]))

    return client.getLinesInSourceFileContents(
        lines_in_files_requested, ttypes.Encoding.BASE64)


def convert_report_data_to_report(
    client,
    reports_data: List[ttypes.ReportData],
    output_formats: List[str]
) -> List[Report]:
    """
    Convert the given report data list to local reports.

    If one of the given output format is HTML it will get full source file
    contents from the server otherwise for performance reason it will
    get only necessarry line contents.
    """
    reports = []

    if not reports_data:
        return reports

    source_line_contents = {}
    if 'html' not in output_formats:
        source_line_contents = get_source_line_contents(client, reports_data)

    file_cache: Dict[int, File] = {}

    def cached_report_file_lookup(file_id: int, file_path: str) -> File:
        """
        It will create a file object with the given 'file_path' attribute if
        file data is not found in the cache already based on the following
        conditions:
          - if HTML output is given it will try to get source file data for the
            given file id from the server and set the content attribute of the
            file object. File content is needed only by the HTML output
            converter.
          - otherwise it will create a file object where file content will not
            be filled automatically but will be loaded from the host machine
            when it is needed.

        Finally, it returns the source file data from the cache.
        """
        nonlocal file_cache

        if file_id not in file_cache:
            file_cache[file_id] = File(file_path, file_id)

            if 'html' in output_formats:
                source = client.getSourceFileData(
                    file_id, True, ttypes.Encoding.BASE64)
                file_cache[file_id].content = convert.from_b64(
                    source.fileContent)

        return file_cache[file_id]

    # Convert reports data to reports.
    for report_data in reports_data:
        report = report_type_converter.to_report(
            report_data, cached_report_file_lookup)

        report.changed_files = []
        report.source_code_comments = []

        if source_line_contents:
            source_line = convert.from_b64(
                source_line_contents[report_data.fileId][report_data.line])
            report.source_line = f"{source_line}{os.linesep}"

        reports.append(report)

    return reports


def get_diff_local_dir_remote_run(
    client,
    report_filter: ttypes.ReportFilter,
    diff_type: ttypes.DiffType,
    output_formats: List[str],
    report_dirs: List[str],
    baseline_files: List[str],
    remote_run_names: List[str]
) -> Tuple[List[Report], List[str], List[str]]:
    """ Compare a local report directory with a remote run. """
    filtered_reports = []
    filtered_report_hashes: Set[str] = set()

    context = webserver_context.get_context()
    report_dir_results = get_report_dir_results(
        report_dirs, report_filter, context.checker_labels)

    suppressed_in_code = \
        get_suppressed_reports(report_dir_results, report_filter.reviewStatus)

    run_ids, run_names, tag_ids = \
        process_run_args(client, remote_run_names)
    local_report_hashes = set([r.report_hash for r in report_dir_results])

    review_status_filter = ttypes.ReviewStatusRuleFilter()
    review_status_filter.reportHashes = local_report_hashes
    review_status_filter.reviewStatuses = [
        ttypes.ReviewStatus.FALSE_POSITIVE,
        ttypes.ReviewStatus.INTENTIONAL]
    closed_hashes = set(
        rule.reportHash for rule in
        client.getReviewStatusRules(
            review_status_filter, None, None, None))

    report_dir_results = [
        r for r in report_dir_results if
        r.review_status.status not in ['false_positive', 'intentional'] and
        (r.report_hash not in closed_hashes or
         r.review_status.status == 'confirmed')]
    local_report_hashes = set(r.report_hash for r in report_dir_results)
    local_report_hashes.update(
        baseline.get_report_hashes(baseline_files) - closed_hashes)

    if diff_type == ttypes.DiffType.NEW:
        # Get report hashes which can be found only in the remote runs.
        remote_hashes = client.getDiffResultsHash(
            run_ids, local_report_hashes, ttypes.DiffType.RESOLVED,
            None, tag_ids)

        results = get_diff_base_results(
            client, report_filter, run_ids, remote_hashes, suppressed_in_code)

        filtered_reports.extend(
            convert_report_data_to_report(client, results, output_formats))
    elif diff_type == ttypes.DiffType.UNRESOLVED:
        # Get remote hashes which can be found in the remote run and in the
        # local report directory.
        remote_hashes = client.getDiffResultsHash(
            run_ids, local_report_hashes, ttypes.DiffType.UNRESOLVED,
            None, tag_ids)

        filtered_report_hashes = local_report_hashes.copy()
        for result in report_dir_results:
            rep_h = result.report_hash
            filtered_report_hashes.discard(rep_h)
            if rep_h in remote_hashes and rep_h not in suppressed_in_code:
                filtered_reports.append(result)
        filtered_report_hashes &= set(remote_hashes)

        # Try to get missing report from the server based on the report
        # hashes.
        if filtered_report_hashes:
            results = get_diff_base_results(
                client, report_filter, run_ids, list(filtered_report_hashes),
                suppressed_in_code)

            for result in results:
                filtered_report_hashes.discard(result.bugHash)

            filtered_reports.extend(
                convert_report_data_to_report(client, results, output_formats))
    elif diff_type == ttypes.DiffType.RESOLVED:
        # Get remote hashes which can be found in the remote run and in the
        # local report directory.
        remote_hashes = client.getDiffResultsHash(
            run_ids, local_report_hashes, ttypes.DiffType.UNRESOLVED,
            None, tag_ids)

        filtered_report_hashes = local_report_hashes.copy()
        for result in report_dir_results:
            filtered_report_hashes.discard(result.report_hash)
            if result.report_hash not in remote_hashes:
                filtered_reports.append(result)
        filtered_report_hashes -= set(remote_hashes)

    return filtered_reports, filtered_report_hashes, run_names


def get_diff_remote_run_local_dir(
    client,
    report_filter: ttypes.ReportFilter,
    diff_type: ttypes.DiffType,
    output_formats: List[str],
    remote_run_names: List[str],
    report_dirs: List[str],
    baseline_files: List[str]
) -> Tuple[List[Report], List[str], List[str]]:
    """ Compares a remote run with a local report directory. """
    filtered_reports = []
    filtered_report_hashes = []

    context = webserver_context.get_context()
    report_dir_results = get_report_dir_results(
        report_dirs, report_filter, context.checker_labels)
    suppressed_in_code = \
        get_suppressed_reports(report_dir_results, report_filter.reviewStatus)

    run_ids, run_names, tag_ids = \
        process_run_args(client, remote_run_names)
    local_report_hashes = set([r.report_hash for r in report_dir_results])

    review_status_filter = ttypes.ReviewStatusRuleFilter()
    review_status_filter.reportHashes = local_report_hashes
    review_status_filter.reviewStatuses = [
        ttypes.ReviewStatus.FALSE_POSITIVE,
        ttypes.ReviewStatus.INTENTIONAL]
    closed_hashes = set(
        rule.reportHash for rule in
        client.getReviewStatusRules(
            review_status_filter, None, None, None))

    report_dir_results = [
        r for r in report_dir_results if
        r.review_status.status not in ['false_positive', 'intentional'] and
        (r.report_hash not in closed_hashes or
         r.review_status.status == 'confirmed')]

    local_report_hashes = set(r.report_hash for r in report_dir_results)
    local_report_hashes.update(
        baseline.get_report_hashes(baseline_files) - closed_hashes)

    remote_hashes = client.getDiffResultsHash(
        run_ids, local_report_hashes, diff_type, None, tag_ids)

    if not remote_hashes:
        return filtered_reports, filtered_report_hashes, run_names

    if diff_type in [ttypes.DiffType.NEW, ttypes.DiffType.UNRESOLVED]:
        # Shows reports from the report dir which are not present in
        # the baseline (NEW reports) or appear in both side (UNRESOLVED
        # reports) and not suppressed in the code.
        filtered_report_hashes = set(remote_hashes)

        for result in report_dir_results:
            rep_h = result.report_hash
            filtered_report_hashes.discard(rep_h)
            if rep_h in remote_hashes and rep_h not in suppressed_in_code:
                filtered_reports.append(result)
    elif diff_type == ttypes.DiffType.RESOLVED:
        # Show bugs in the baseline (server) which are not present in
        # the report dir or suppressed.
        results = get_diff_base_results(
            client, report_filter, run_ids, remote_hashes, suppressed_in_code)

        filtered_reports.extend(
            convert_report_data_to_report(client, results, output_formats))

    return filtered_reports, filtered_report_hashes, run_names


def get_diff_remote_runs(
    client,
    report_filter: ttypes.ReportFilter,
    diff_type: ttypes.DiffType,
    output_formats: List[str],
    remote_base_run_names: Iterable[str],
    remote_new_run_names: Iterable[str]
) -> Tuple[List[Report], List[str], List[str]]:
    """
    Compares two remote runs and returns the filtered results.
    """
    base_ids, base_run_names, base_run_tags = \
        process_run_args(client, remote_base_run_names)
    report_filter.runTag = base_run_tags

    cmp_data = ttypes.CompareData()
    cmp_data.diffType = diff_type

    new_ids, new_run_names, new_run_tags = \
        process_run_args(client, remote_new_run_names)
    cmp_data.runIds = new_ids
    cmp_data.runTag = new_run_tags

    sort_mode = [(ttypes.SortMode(
        ttypes.SortType.FILENAME,
        ttypes.Order.ASC))]

    # Generally speaking, we only compare outstanding reports in both the
    # baseline and newline analyses, so it wouldn't make sense to show reports
    # whose detection status indicates otherwise. However, for tags, we check
    # whether a report is outstanding at the time of the tag, so in the context
    # of that tag, a resolved report may actually be outstanding, so we clear
    # the detection status filter.
    # TODO: We don't support diffs with timestamps on the cmdline, but we
    # should do this for them as well.
    if base_run_tags or new_run_tags:
        # If the value is not the default, lets warn the user that we overwrote
        # it.
        default_detection_status = [ttypes.DetectionStatus.NEW,
                                    ttypes.DetectionStatus.REOPENED,
                                    ttypes.DetectionStatus.UNRESOLVED]
        if report_filter.detectionStatus != default_detection_status and \
           report_filter.detectionStatus != []:
            LOG.warning("--detection-status is ignored when comparing tags, "
                        "showing reports regardless of detection status.")

        report_filter.detectionStatus = []

    all_results = get_run_results(
        client, base_ids, constants.MAX_QUERY_SIZE, 0, sort_mode,
        report_filter, cmp_data, True)

    reports = \
        convert_report_data_to_report(client, all_results, output_formats)
    return reports, base_run_names, new_run_names


def get_diff_local_dirs(
    report_filter: ttypes.ReportFilter,
    diff_type: ttypes.DiffType,
    # TODO: No output_format argument? How come? Maybe the other functions
    # don't need it either?
    report_dirs: List[str],
    baseline_files: List[str],
    new_report_dirs: List[str],
    new_baseline_files: List[str]
) -> Tuple[List[Report], List[str]]:
    """
    Compares two report directories and returns the filtered results.
    """
    filtered_reports = []
    filtered_report_hashes = []

    context = webserver_context.get_context()
    statuses_str = [ttypes.ReviewStatus._VALUES_TO_NAMES[x].lower()
                    for x in report_filter.reviewStatus]
    statuses_str.append('unreviewed')

    base_results = get_report_dir_results(
        report_dirs, report_filter, context.checker_labels)
    base_results = [res for res in base_results
                    if res.review_status.status in statuses_str]

    new_results = get_report_dir_results(
        new_report_dirs, report_filter, context.checker_labels)
    new_results = [res for res in new_results
                   if res.review_status.status in statuses_str]

    base_hashes = set([res.report_hash for res in base_results])
    new_hashes = set([res.report_hash for res in new_results])

    # Add hashes from the baseline files.
    base_hashes.update(baseline.get_report_hashes(baseline_files))
    new_hashes.update(baseline.get_report_hashes(new_baseline_files))

    if diff_type == ttypes.DiffType.NEW:
        filtered_report_hashes = new_hashes.copy()
        for res in new_results:
            filtered_report_hashes.discard(res.report_hash)

            if res.report_hash not in base_hashes:
                filtered_reports.append(res)
    if diff_type == ttypes.DiffType.UNRESOLVED:
        filtered_report_hashes = new_hashes.copy()
        for res in new_results:
            filtered_report_hashes.discard(res.report_hash)

            if res.report_hash in base_hashes:
                filtered_reports.append(res)
    elif diff_type == ttypes.DiffType.RESOLVED:
        filtered_report_hashes = base_hashes.copy()
        for res in base_results:
            filtered_report_hashes.discard(res.report_hash)

            if res.report_hash not in new_hashes:
                filtered_reports.append(res)

    return filtered_reports, filtered_report_hashes


def print_reports(
    print_steps: bool,
    reports: List[Report],
    report_hashes: Iterable[str],
    output_dir: str,
    output_formats: List[str]
):
    if report_hashes:
        LOG.info("Couldn't get local reports for the following baseline "
                 "report hashes: %s", ', '.join(sorted(report_hashes)))

    statistics = Statistics()
    changed_files: Set[str] = set()
    html_builder: Optional[report_to_html.HtmlBuilder] = None
    for report in reports:
        statistics.add_report(report)

        if report.changed_files:
            changed_files.update(report.changed_files)

        repo_dir = os.environ.get('CC_REPO_DIR')
        if repo_dir:
            report.trim_path_prefixes([repo_dir])

    processed_file_paths = set()
    for output_format in output_formats:
        if output_format == 'plaintext':
            file_report_map = plaintext.get_file_report_map(reports)
            plaintext.convert(
                ReviewStatusHandler(),
                file_report_map, processed_file_paths, print_steps)

        context = webserver_context.get_context()
        if output_format == 'html':
            if not html_builder:
                html_builder = report_to_html.HtmlBuilder(
                    context.path_plist_to_html_dist,
                    context.checker_labels)

            file_report_map = plaintext.get_file_report_map(reports)

            LOG.info("Generating HTML output files to file://%s "
                     "directory:", output_dir)

            for file_path, file_reports in file_report_map.items():
                file_name = os.path.basename(file_path)
                h = int(
                    hashlib.md5(
                        file_path.encode('utf-8')).hexdigest(),
                    16) % (10 ** 8)

                output_file_path = os.path.join(
                    output_dir, f"{file_name}_ {str(h)}.html")
                html_builder.create(output_file_path, file_reports)

        if output_format in ['csv', 'rows', 'table']:
            header = ['File', 'Checker', 'Severity', 'Msg', 'Source']
            rows = []
            for report in reports:
                if report.source_line is None:
                    continue

                checked_file = f"{report.file.path}:{str(report.line)}:" \
                               f"{str(report.column)}"
                rows.append((
                    report.severity,
                    checked_file,
                    report.message,
                    report.checker_name,
                    report.source_line.rstrip()))

            print(twodim.to_str(output_format, header, rows))

        if output_format == 'json':
            data = report_to_json.convert(reports)

            report_json = os.path.join(output_dir, 'reports.json') \
                if output_dir else None
            dump_json_output(data, report_json)

        if output_format == 'gerrit':
            data = gerrit.convert(reports)

            report_json = os.path.join(
                output_dir, 'gerrit_review.json') if output_dir else None
            dump_json_output(data, report_json)

        if output_format == 'codeclimate':
            data = codeclimate.convert(reports)

            report_json = os.path.join(
                output_dir, 'codeclimate_issues.json') \
                if output_dir else None
            dump_json_output(data, report_json)

    if 'html' in output_formats:
        html_builder.finish(output_dir, statistics)

    if 'plaintext' in output_formats:
        statistics.write()

    reports_helper.dump_changed_files(changed_files)


def handle_diff_results(args):
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    output_formats = args.output_format
    if output_formats != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    output_dir = args.export_dir if 'export_dir' in args else None
    if len(output_formats) > 1 and ('export_dir' not in args):
        LOG.error("Export directory is required if multiple output formats "
                  "are selected!")
        sys.exit(1)

    if 'html' in output_formats and not output_dir:
        LOG.error("Argument --export not allowed without argument --output "
                  "when exporting to HTML.")
        sys.exit(1)

    if 'gerrit' in output_formats and \
            not gerrit.mandatory_env_var_is_set():
        sys.exit(1)

    check_deprecated_arg_usage(args)

    if 'clean' in args and os.path.isdir(output_dir):
        print("Previous analysis results in '{0}' have been removed, "
              "overwriting with current results.".format(output_dir))
        shutil.rmtree(output_dir)

    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    client = None

    basename_local_dirs, basename_baseline_files, basename_run_names = \
        filter_local_file_remote_run(args.base_run_names)

    newname_local_dirs, newname_baseline_files, newname_run_names = \
        filter_local_file_remote_run(args.new_run_names)

    has_different_run_args = False
    if (basename_local_dirs or basename_baseline_files) and basename_run_names:
        LOG.error("All base run names must have the same type: local "
                  "directory (%s) / baseline files (%s) or run names (%s).",
                  ', '.join(basename_local_dirs),
                  ', '.join(basename_baseline_files),
                  ', '.join(basename_run_names))
        has_different_run_args = True

    if newname_local_dirs and newname_run_names:
        LOG.error("All new run names must have the same type: local "
                  "directory (%s) / baseline files (%s) or run names (%s).",
                  ', '.join(newname_local_dirs),
                  ', '.join(newname_baseline_files),
                  ', '.join(newname_run_names))
        has_different_run_args = True

    if has_different_run_args:
        sys.exit(1)

    if basename_local_dirs:
        LOG.info("Matching local report directories (--baseline): %s",
                 ', '.join(basename_local_dirs))
    if basename_baseline_files:
        LOG.info("Matching local baseline files (--baseline): %s",
                 ', '.join(basename_baseline_files))

    if newname_local_dirs:
        LOG.info("Matching local report directories (--newname): %s",
                 ', '.join(newname_local_dirs))
    if newname_baseline_files:
        LOG.info("Matching local baseline files (--newname): %s",
                 ', '.join(newname_baseline_files))

    # We set up the client if we are not comparing two local report directories
    # or baseline files.
    if basename_run_names or newname_run_names:
        if basename_run_names:
            LOG.info("Given remote runs (--baseline): %s",
                     ', '.join(basename_run_names))
        if newname_run_names:
            LOG.info("Given remote runs (--newname): %s",
                     ', '.join(newname_run_names))

        try:
            client = setup_client(args.product_url)
        except SystemExit as sexit:
            LOG.error("Failed to get remote runs from server! Please "
                      "make sure that CodeChecker server is running on host "
                      "'%s' or these are valid directory paths!",
                      args.product_url)
            raise sexit

    if 'component' in args:
        check_existing_source_components(client, args.component)

    print_steps = 'print_steps' in args
    report_hashes = []
    if (basename_local_dirs or basename_baseline_files) and \
       (newname_local_dirs or newname_baseline_files):
        report_filter = parse_report_filter_offline(args)
        diff_type = get_diff_type(args)

        reports, report_hashes = get_diff_local_dirs(
            report_filter, diff_type,
            basename_local_dirs, basename_baseline_files,
            newname_local_dirs, newname_baseline_files)

        print_reports(print_steps, reports, report_hashes, output_dir,
                      output_formats)
        LOG.info("Compared the following local files / directories: %s and %s",
                 ', '.join([*basename_local_dirs, *basename_baseline_files]),
                 ', '.join([*newname_local_dirs, *newname_baseline_files]))
    elif newname_local_dirs or newname_baseline_files:
        report_filter = parse_report_filter(client, args)
        diff_type = get_diff_type(args)

        reports, report_hashes, matching_base_run_names = \
            get_diff_remote_run_local_dir(
                client, report_filter, diff_type, output_formats,
                basename_run_names,
                newname_local_dirs, newname_baseline_files)

        print_reports(print_steps, reports, report_hashes, output_dir,
                      output_formats)
        LOG.info("Compared remote run(s) %s (matching: %s) and local files / "
                 "report directory(s) %s",
                 ', '.join(basename_run_names),
                 ', '.join(matching_base_run_names),
                 ', '.join([*newname_local_dirs, *newname_baseline_files]))
    elif (basename_local_dirs or basename_baseline_files):
        report_filter = parse_report_filter(client, args)
        diff_type = get_diff_type(args)

        reports, report_hashes, matching_new_run_names = \
            get_diff_local_dir_remote_run(
                client, report_filter, diff_type, output_formats,
                basename_local_dirs,
                basename_baseline_files, newname_run_names)

        print_reports(print_steps, reports, report_hashes, output_dir,
                      output_formats)
        LOG.info("Compared local files / report directory(s) %s and remote "
                 "run(s) %s (matching: %s).",
                 ', '.join([*basename_local_dirs, *basename_baseline_files]),
                 ', '.join(newname_run_names),
                 ', '.join(matching_new_run_names))
    else:
        report_filter = parse_report_filter(client, args)
        diff_type = get_diff_type(args)

        reports, matching_base_run_names, matching_new_run_names = \
            get_diff_remote_runs(
                client, report_filter, diff_type, output_formats,
                basename_run_names, newname_run_names)
        print_reports(print_steps, reports, None, output_dir, output_formats)
        LOG.info("Compared multiple remote runs %s (matching: %s) and %s "
                 "(matching: %s)",
                 ', '.join(basename_run_names),
                 ', '.join(matching_base_run_names),
                 ', '.join(newname_run_names),
                 ', '.join(matching_new_run_names))

    if len(reports) != 0 or len(report_hashes) != 0:
        sys.exit(2)


def handle_list_result_types(args):
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args and args.output_format != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)
    check_deprecated_arg_usage(args)

    def get_statistics(
        client, run_ids, field, values,
        all_checkers_report_filter
    ):
        report_filter = deepcopy(all_checkers_report_filter)

        setattr(report_filter, field, values)
        checkers = client.getCheckerCounts(run_ids,
                                           report_filter,
                                           None,
                                           None,
                                           0)

        return dict((res.name, res.count) for res in checkers)

    def checker_count(checker_dict, key):
        return checker_dict.get(key, 0)

    client = setup_client(args.product_url)

    if 'component' in args:
        check_existing_source_components(client, args.component)

    run_ids = None
    if 'all_results' not in args:
        run_filter = ttypes.RunFilter(names=args.names)
        run_ids = [run.runId for run in get_run_data(client, run_filter)]
        if not run_ids:
            LOG.warning("No runs were found!")
            sys.exit(1)

    all_checkers_report_filter = parse_report_filter(client, args)

    all_checkers = client.getCheckerCounts(run_ids,
                                           all_checkers_report_filter,
                                           None,
                                           None,
                                           0)
    all_checkers_dict = dict((res.name, res) for res in all_checkers)

    unrev_checkers = get_statistics(client, run_ids, 'reviewStatus',
                                    [ttypes.ReviewStatus.UNREVIEWED],
                                    all_checkers_report_filter)

    confirmed_checkers = get_statistics(client, run_ids, 'reviewStatus',
                                        [ttypes.ReviewStatus.CONFIRMED],
                                        all_checkers_report_filter)

    false_checkers = get_statistics(client, run_ids, 'reviewStatus',
                                    [ttypes.ReviewStatus.FALSE_POSITIVE],
                                    all_checkers_report_filter)

    intentional_checkers = get_statistics(client, run_ids, 'reviewStatus',
                                          [ttypes.ReviewStatus.INTENTIONAL],
                                          all_checkers_report_filter)

    # Get severity counts
    sev_count = client.getSeverityCounts(
        run_ids, all_checkers_report_filter, None)

    severities = []
    severity_total = 0
    for key, count in sorted(list(sev_count.items()),
                             reverse=True):
        severities.append(dict(
            severity=ttypes.Severity._VALUES_TO_NAMES[key],
            reports=count))
        severity_total += count

    all_results = []
    total = defaultdict(int)
    for key, checker_data in sorted(list(all_checkers_dict.items()),
                                    key=lambda x: x[1].severity,
                                    reverse=True):
        all_results.append(dict(
            checker=key,
            severity=ttypes.Severity._VALUES_TO_NAMES[checker_data.severity],
            reports=checker_data.count,
            unreviewed=checker_count(unrev_checkers, key),
            confirmed=checker_count(confirmed_checkers, key),
            false_positive=checker_count(false_checkers, key),
            intentional=checker_count(intentional_checkers, key)
        ))
        total['total_reports'] += checker_data.count
        total['total_unreviewed'] += checker_count(unrev_checkers, key)
        total['total_confirmed'] += checker_count(confirmed_checkers, key)
        total['total_false_positive'] += checker_count(false_checkers, key)
        total['total_intentional'] += checker_count(intentional_checkers, key)

    if args.output_format == 'json':
        print(CmdLineOutputEncoder().encode(all_results))
    else:
        header = ['Checker', 'Severity', 'Unreviewed', 'Confirmed',
                  'False positive', 'Intentional', 'All reports']

        rows = []
        for stat in all_results:
            rows.append((stat['checker'],
                         stat['severity'],
                         str(stat['unreviewed']),
                         str(stat['confirmed']),
                         str(stat['false_positive']),
                         str(stat['intentional']),
                         str(stat['reports'])))

        rows.append(('Total', '-',
                     str(total['total_unreviewed']),
                     str(total['total_confirmed']),
                     str(total['total_false_positive']),
                     str(total['total_intentional']),
                     str(total['total_reports'])))

        print(twodim.to_str(args.output_format, header, rows,
                            separate_footer=True))

        # Print severity counts
        header = ['Severity', 'All reports']

        rows = []
        for stat in severities:
            rows.append((stat['severity'],
                         str(stat['reports'])))

        rows.append(('Total', str(severity_total)))

        print(twodim.to_str(args.output_format, header, rows,
                            separate_footer=True))


def handle_remove_run_results(args):

    init_logger(args.verbose if 'verbose' in args else None)

    client = setup_client(args.product_url)

    run_filter = process_run_filter_conditions(args)
    if client.removeRun(None, run_filter):
        LOG.info("Done.")
    else:
        LOG.warning(
            "No runs were removed, check if the given criteria match existing "
            "run names.")


def handle_update_run(args):
    """
    Argument handler for the 'CodeChecker cmd update' subcommand.
    """
    init_logger(args.verbose if 'verbose' in args else None)

    if not args.new_run_name:
        LOG.error("The new run name can not be empty!")
        sys.exit(1)

    client = setup_client(args.product_url)

    run_info = check_run_names(client, [args.run_name])
    run = run_info.get(args.run_name)
    if not run:
        LOG.warning("No run name with the name '%s' was found.", args.name)
        return

    try:
        client.updateRunData(run.runId, args.new_run_name)
    except RequestFailed as reqfail:
        LOG.error(reqfail.message)
        sys.exit(1)

    LOG.info("Done.")


def handle_suppress(args):

    init_logger(args.verbose if 'verbose' in args else None)

    limit = constants.MAX_QUERY_SIZE

    client = setup_client(args.product_url)

    run_info = check_run_names(client, [args.name])
    run = run_info.get(args.name)
    if not run:
        LOG.warning("No run name with the name '%s' was found.", args.name)
        return

    if 'input' in args:
        with open(args.input, encoding='utf-8', errors='ignore') as supp_file:
            suppress_data = suppress_file_handler.get_suppress_data(supp_file)

        for bug_id, file_name, comment, status in suppress_data:
            file_name = '%' + file_name
            bug_hash_filter = ttypes.ReportFilter(filepath=[file_name],
                                                  reportHash=[bug_id])
            reports = client.getRunResults([run.runId], limit, 0, None,
                                           bug_hash_filter,
                                           None, False)

            for report in reports:
                rw_status = ttypes.ReviewStatus.FALSE_POSITIVE
                if status == 'confirmed':
                    rw_status = ttypes.ReviewStatus.CONFIRMED
                elif status == 'intentional':
                    rw_status = ttypes.ReviewStatus.INTENTIONAL

                client.changeReviewStatus(report.reportId, rw_status, comment)


def handle_login(args):

    init_logger(args.verbose if 'verbose' in args else None)

    protocol, host, port = split_server_url(args.server_url)
    login_user(protocol, host, port, args.username,
               login='logout' not in args)


def handle_list_run_histories(args):
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args and args.output_format != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    client = setup_client(args.product_url)
    run_ids = None
    if 'names' in args:
        run_filter = ttypes.RunFilter(names=args.names)
        runs = get_run_data(client, run_filter)
        run_ids = [r.runId for r in runs]

    run_history = client.getRunHistory(run_ids, None, None, None)

    if args.output_format == 'json':
        print(CmdLineOutputEncoder().encode(run_history))
    else:  # plaintext, csv
        header = ['Date', 'Run name', 'Version tag', 'User',
                  'CodeChecker version', 'Analyzer statistics', 'Description']
        rows = []
        for h in run_history:
            analyzer_statistics = []
            for analyzer in h.analyzerStatistics:
                stat = h.analyzerStatistics[analyzer]
                num_of_all_files = stat.successful + stat.failed
                analyzer_statistics.append(analyzer + ' (' +
                                           str(num_of_all_files) + '/' +
                                           str(stat.successful) + ')')

            rows.append((h.time,
                         h.runName,
                         h.versionTag if h.versionTag else '',
                         h.user,
                         h.codeCheckerVersion if h.codeCheckerVersion else '',
                         ', '.join(analyzer_statistics),
                         h.description if h.description else ''))

        print(twodim.to_str(args.output_format, header, rows))


def handle_export(args):

    stream = 'stderr'
    init_logger(args.verbose if 'verbose' in args else None, stream)

    client = setup_client(args.product_url)
    run_filter = process_run_filter_conditions(args)

    runs = get_run_data(client, run_filter)
    if not runs:
        LOG.warning("No runs found")
        sys.exit(1)

    report = client.exportData(run_filter)
    print(CmdLineOutputEncoder().encode(report))


def handle_import(args):
    """
    Argument handler for the 'CodeChecker cmd import' subcommand
    """

    init_logger(args.verbose if 'verbose' in args else None)

    client = setup_client(args.product_url)

    data = load_json(args.input, default=None)
    if not data:
        LOG.error("Failed to import data!")
        sys.exit(1)

    # Convert Json comments into CommentDataList
    comment_data_list = defaultdict(list)
    for bug_hash, comments in data['comments'].items():
        for comment in comments:
            comment_data = ttypes.CommentData(
                id=comment['id'],
                author=comment['author'],
                message=comment['message'],
                createdAt=comment['createdAt'],
                kind=comment['kind'])
            comment_data_list[bug_hash].append(comment_data)

    # Convert Json reviews into ReviewData
    review_data_list = {}
    for bug_hash, review in data['reviewData'].items():
        review_data_list[bug_hash] = ttypes.ReviewData(
            status=review['status'],
            comment=review['comment'],
            author=review['author'],
            date=review['date'])

    exportData = ttypes.ExportData(comments=comment_data_list,
                                   reviewData=review_data_list)

    status = client.importData(exportData)
    if not status:
        LOG.error("Failed to import data!")
        sys.exit(1)

    LOG.info("Import successful!")



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
""" JSON output helpers. """

from typing import Dict, List

from codechecker_report_converter.report import Report


def convert(reports: List[Report]) -> Dict:
    """ Convert the given reports to JSON format. """
    version = 1

    json_reports = []
    for report in reports:
        json_reports.append(report.to_json())

    return {"version": version, "reports": json_reports}



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Parser for PostgreSQL libpq password file.
"""


def _match_field(line, field):
    """
    Return the remaining part of the line in case of matching field,
    otherwise it returns None.
    """

    if line is None or len(line) < 2:
        return None

    if line[0] == '*' and line[1] == ':':
        # Match (star).
        return line[2:]

    escaped = False
    while line:
        if not escaped and line[0] == '\\':
            line = line[1:]
            escaped = True

        if not line:
            return None

        if not escaped and line[0] == ':' and not field:
            # match
            return line[1:]

        escaped = False
        if not field:
            return None
        elif field[0] == line[0]:
            line = line[1:]
            field = field[1:]
        else:
            return None
    return None


def _match_line(line, hostname, port, database, username):
    """
    Tries to match the given line to the given hostname, port, database,
    and username.

    Returns non None on match.
    """

    pw = _match_field(line, hostname)
    pw = _match_field(pw, port)
    pw = _match_field(pw, database)
    pw = _match_field(pw, username)
    if pw is None:
        return pw

    # The password is still escaped.
    escaped = False
    password = ''
    for c in pw:
        if not escaped and c == '\\':
            escaped = True
            continue

        escaped = False
        password += c
    return password


def get_password_from_file(passfile_path, hostname, port, database, username):
    """
    Parser for PostgreSQL libpq password file.

    Returns None on no matching entry, otherwise it returns the password.

    For file format see:
    http://www.postgresql.org/docs/current/static/libpq-pgpass.html
    """

    if not hostname or not port or not database or not username:
        return None

    with open(passfile_path, 'r',
              encoding="utf-8", errors="ignore") as passfile:
        for line in passfile:
            pw = _match_line(line.strip(), hostname, port, database, username)
            if pw:
                return pw

    return None



"""
New table for server config

Revision ID: dec6feb991e6
Revises:     3335ff7593cc
Create Date: 2019-02-05 17:36:46.527079
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'dec6feb991e6'
down_revision = '3335ff7593cc'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'server_configurations',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('config_key', sa.String(), nullable=False),
        sa.Column('config_value', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_server_configurations'))
    )


def downgrade():
    op.drop_table('server_configurations')



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class ErrorCode(object):
    DATABASE = 0
    IOERROR = 1
    GENERAL = 2
    AUTH_DENIED = 3
    UNAUTHORIZED = 4
    API_MISMATCH = 5
    SOURCE_FILE = 6
    REPORT_FORMAT = 7

    _VALUES_TO_NAMES = {
        0: "DATABASE",
        1: "IOERROR",
        2: "GENERAL",
        3: "AUTH_DENIED",
        4: "UNAUTHORIZED",
        5: "API_MISMATCH",
        6: "SOURCE_FILE",
        7: "REPORT_FORMAT",
    }

    _NAMES_TO_VALUES = {
        "DATABASE": 0,
        "IOERROR": 1,
        "GENERAL": 2,
        "AUTH_DENIED": 3,
        "UNAUTHORIZED": 4,
        "API_MISMATCH": 5,
        "SOURCE_FILE": 6,
        "REPORT_FORMAT": 7,
    }


class Permission(object):
    """
    The following permission scopes exist.

    SYSTEM: These permissions are global to the running CodeChecker server.
      In this case, the 'extraParams' field is empty.

    PRODUCT: These permissions are configured per-product.
      The extra data field looks like the following object:
        { i64 productID }
    """
    SUPERUSER = 1
    PERMISSION_VIEW = 2
    PRODUCT_ADMIN = 16
    PRODUCT_ACCESS = 17
    PRODUCT_STORE = 18
    PRODUCT_VIEW = 19

    _VALUES_TO_NAMES = {
        1: "SUPERUSER",
        2: "PERMISSION_VIEW",
        16: "PRODUCT_ADMIN",
        17: "PRODUCT_ACCESS",
        18: "PRODUCT_STORE",
        19: "PRODUCT_VIEW",
    }

    _NAMES_TO_VALUES = {
        "SUPERUSER": 1,
        "PERMISSION_VIEW": 2,
        "PRODUCT_ADMIN": 16,
        "PRODUCT_ACCESS": 17,
        "PRODUCT_STORE": 18,
        "PRODUCT_VIEW": 19,
    }


class DBStatus(object):
    """
    Status information about the database backend.
    """
    OK = 0
    MISSING = 1
    FAILED_TO_CONNECT = 2
    SCHEMA_MISMATCH_OK = 3
    SCHEMA_MISMATCH_NO = 4
    SCHEMA_MISSING = 5
    SCHEMA_INIT_ERROR = 6
    SCHEMA_UPGRADE_FAILED = 7

    _VALUES_TO_NAMES = {
        0: "OK",
        1: "MISSING",
        2: "FAILED_TO_CONNECT",
        3: "SCHEMA_MISMATCH_OK",
        4: "SCHEMA_MISMATCH_NO",
        5: "SCHEMA_MISSING",
        6: "SCHEMA_INIT_ERROR",
        7: "SCHEMA_UPGRADE_FAILED",
    }

    _NAMES_TO_VALUES = {
        "OK": 0,
        "MISSING": 1,
        "FAILED_TO_CONNECT": 2,
        "SCHEMA_MISMATCH_OK": 3,
        "SCHEMA_MISMATCH_NO": 4,
        "SCHEMA_MISSING": 5,
        "SCHEMA_INIT_ERROR": 6,
        "SCHEMA_UPGRADE_FAILED": 7,
    }


class RequestFailed(TException):
    """
    Attributes:
     - errorCode
     - message
     - extraInfo
    """


    def __init__(self, errorCode=None, message=None, extraInfo=None,):
        self.errorCode = errorCode
        self.message = message
        self.extraInfo = extraInfo

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.errorCode = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.extraInfo = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.extraInfo.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('RequestFailed')
        if self.errorCode is not None:
            oprot.writeFieldBegin('errorCode', TType.I32, 1)
            oprot.writeI32(self.errorCode)
            oprot.writeFieldEnd()
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 2)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        if self.extraInfo is not None:
            oprot.writeFieldBegin('extraInfo', TType.LIST, 3)
            oprot.writeListBegin(TType.STRING, len(self.extraInfo))
            for iter6 in self.extraInfo:
                oprot.writeString(iter6.encode('utf-8') if sys.version_info[0] == 2 else iter6)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(RequestFailed)
RequestFailed.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'errorCode', None, None, ),  # 1
    (2, TType.STRING, 'message', 'UTF8', None, ),  # 2
    (3, TType.LIST, 'extraInfo', (TType.STRING, 'UTF8', False), None, ),  # 3
)
fix_spec(all_structs)
del all_structs



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


import argparse
import logging
import os
import sys

# If we run this script in an environment where
# 'codechecker_merge_clang_extdef_mappings' module is not available we should
# add the grandparent directory of this file to the system path.
# TODO: This section will not be needed when CodeChecker will be delivered as
# a python package and will be installed in a virtual environment with all the
# dependencies.
if __name__ == '__main__':
    current_dir = os.path.dirname(os.path.realpath(__file__))
    os.sys.path.append(os.path.dirname(current_dir))

from codechecker_merge_clang_extdef_mappings import \
    merge_clang_extdef_mappings  # noqa


LOG = logging.getLogger('MergeClangExtdefMappings')

msg_formatter = logging.Formatter('[%(levelname)s] - %(message)s')
log_handler = logging.StreamHandler(sys.stdout)
log_handler.setFormatter(msg_formatter)
LOG.setLevel(logging.INFO)
LOG.addHandler(log_handler)


def __add_arguments_to_parser(parser):
    """ Add arguments to the the given parser. """
    parser.add_argument('-i', '--input',
                        type=str,
                        metavar='input',
                        required=True,
                        help="Folder which contains multiple output of the "
                             "'clang-extdef-mapping' tool.")

    parser.add_argument('-o', '--output',
                        type=str,
                        metavar='output',
                        required=True,
                        help="Output file where the merged function maps will "
                             "be stored into.")


def main():
    """ Merge CTU funcs maps main command line. """
    parser = argparse.ArgumentParser(
        prog="merge-clang-extdef-mappings",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Merge individual clang extdef mapping files into one "
                    "mapping file.",
        epilog="""Example:
  merge-clang-extdef-mappings -i /path/to/fn_map_folder
  /path/to/externalDefMap.txt""")

    __add_arguments_to_parser(parser)

    args = parser.parse_args()

    merge_clang_extdef_mappings.merge(args.input, args.output)


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Clang Static analyzer configuration handler.
"""
from codechecker_common.logger import get_logger

from .. import config_handler

LOG = get_logger('analyzer.clangsa')


class ClangSAConfigHandler(config_handler.AnalyzerConfigHandler):
    """
    Configuration handler for the clang static analyzer.
    """

    def __init__(self, environ):
        super(ClangSAConfigHandler, self).__init__()
        self.ctu_dir = ''
        self.ctu_on_demand = False
        self.enable_z3 = False
        self.enable_z3_refutation = False
        self.environ = environ



"""
Git blame info

Revision ID: ad2a567e513a
Revises:     f8291ab1d6be
Create Date: 2020-12-17 18:08:50.322381
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'ad2a567e513a'
down_revision = 'f8291ab1d6be'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('files',
                  sa.Column('remote_url', sa.String(), nullable=True))
    op.add_column('files',
                  sa.Column('tracking_branch', sa.String(), nullable=True))

    op.add_column('file_contents',
                  sa.Column('blame_info', sa.Binary(), nullable=True))


def downgrade():
    op.drop_column('file_contents', 'blame_info')
    op.drop_column('files', 'remote_url')
    op.drop_column('files', 'tracking_branch')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handles the management of stored user credentials and currently known session
tokens.
"""

import json
import os
import re
import stat

import portalocker

from codechecker_common.logger import get_logger
from codechecker_common.util import load_json

from codechecker_web.shared.env import check_file_owner_rw, get_password_file,\
    get_session_file
from codechecker_web.shared.version import SESSION_COOKIE_NAME as _SCN

LOG = get_logger('system')
SESSION_COOKIE_NAME = _SCN


def simplify_credentials(credentials):
    """ Replace the key of the entries with host:port values.

    Returns a new dictionary from the credentials where the key will be
    replaced by 'host:port' values so protocols and product names will be
    removed if these are given.
    """
    host_entry_pattern = re.compile(
        r'^(?P<protocol>http[s]?://)*(?P<host>[^:\/\s]+):*(?P<port>\d+)*'
        r'/*(?P<product>\w+)*$')

    ret = {}
    for host_entry, auth_string in credentials.items():
        match = host_entry_pattern.match(host_entry)
        if not match:
            LOG.warning("Not a valid host entry: %s.", host_entry)
            continue

        host = match.group('host')
        port = match.group('port')
        host_port = '{0}:{1}'.format(host, port) if port else host

        ret[host_port] = auth_string

    return ret


class UserCredentials:

    def __init__(self):
        LOG.debug("Loading clientside session config.")

        # Check whether user's configuration exists.
        session_cfg_file = get_password_file()
        LOG.info("Checking local passwords or tokens in %s", session_cfg_file)

        scfg_dict = {}

        user_home = os.path.expanduser("~")
        mistyped_cfg_file = os.path.join(user_home,
                                         ".codechecker.password.json")

        if os.path.exists(session_cfg_file):
            check_file_owner_rw(session_cfg_file)
            scfg_dict = load_json(session_cfg_file, {})
            scfg_dict['credentials'] = \
                simplify_credentials(scfg_dict['credentials'])
            if not scfg_dict['credentials']:
                LOG.info("No saved tokens.")
            else:
                LOG.debug("Tokens or passwords were found for these hosts:")
                for k, v in scfg_dict['credentials'].items():
                    user, _ = v.split(":")
                    LOG.debug("  user '%s' host '%s'", user, k)
        elif os.path.exists(mistyped_cfg_file):
            LOG.warning("Typo in file name! Rename '%s' to '%s'.",
                        mistyped_cfg_file, session_cfg_file)
        else:
            LOG.info("Password file not found.")

        if not scfg_dict.get('credentials'):
            scfg_dict['credentials'] = {}

        self.__save = scfg_dict
        self.__autologin = scfg_dict.get('client_autologin', True)
        # Check and load token storage for user.
        self.token_file = get_session_file()
        LOG.info("Checking for local valid sessions.")

        with open(self.token_file, 'a+',
                  encoding="utf-8", errors="ignore") as f:
            f.seek(0)

            try:
                portalocker.lock(f, portalocker.LOCK_EX)

                token_dict = json.loads(f.read())

                check_file_owner_rw(self.token_file)

                self.__tokens = token_dict.get('tokens', {})
                LOG.debug("Found session information for these hosts:")
                for k, _ in self.__tokens.items():
                    LOG.debug("  %s", k)
            except json.JSONDecodeError:
                json.dump({'tokens': {}}, f)
                os.chmod(self.token_file, stat.S_IRUSR | stat.S_IWUSR)

                self.__tokens = {}
            finally:
                portalocker.unlock(f)

    def is_autologin_enabled(self):
        return self.__autologin

    def get_token(self, host, port):
        return self.__tokens.get("{0}:{1}".format(host, port))

    def get_auth_string(self, host, port):
        ret = self.__save['credentials'].get('{0}:{1}'.format(host, port))
        if not ret:
            ret = self.__save['credentials'].get(host)
        if not ret:
            ret = self.__save['credentials'].get('*:{0}'.format(port))
        if not ret:
            ret = self.__save['credentials'].get('*')

        return ret

    def save_token(self, host, port, token, destroy=False):
        if destroy:
            self.__tokens.pop('{0}:{1}'.format(host, port), None)
        else:
            self.__tokens['{0}:{1}'.format(host, port)] = token

        with open(self.token_file, 'w',
                  encoding="utf-8", errors="ignore") as scfg:
            portalocker.lock(scfg, portalocker.LOCK_EX)
            json.dump({'tokens': self.__tokens}, scfg,
                      indent=2, sort_keys=True)
            portalocker.unlock(scfg)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Cppcheck related functions.
"""

from collections import defaultdict
# TODO distutils will be removed in python3.12
from distutils.version import StrictVersion
from pathlib import Path
import os
import pickle
import re
import shlex
import shutil
import subprocess
import xml.etree.ElementTree as ET

from codechecker_common.logger import get_logger

from codechecker_analyzer import analyzer_context, env
from codechecker_analyzer.env import get_binary_in_path

from .. import analyzer_base

from .config_handler import CppcheckConfigHandler
from .result_handler import CppcheckResultHandler

from ..config_handler import CheckerState

LOG = get_logger('analyzer.cppcheck')


def parse_checkers(cppcheck_output):
    """
    Parse cppcheck checkers list given by '--errorlist' flag. Return a list of
    (checker_name, description) pairs.
    """
    checkers = []

    tree = ET.ElementTree(ET.fromstring(cppcheck_output))
    root = tree.getroot()
    errors = root.find('errors')
    for error in errors.findall('error'):
        name = error.attrib.get('id')
        if name:
            name = "cppcheck-" + name
        msg = str(error.attrib.get('msg') or '')
        # TODO: Check severity handling in cppcheck
        # severity = error.attrib.get('severity')

        # checkers.append((name, msg, severity))
        checkers.append((name, msg))

    return checkers


def parse_version(cppcheck_output):
    """
    Parse cppcheck version output and return the version number.
    """
    version_re = re.compile(r'^Cppcheck(.*?)(?P<version>[\d\.]+)')
    match = version_re.match(cppcheck_output)
    if match:
        return match.group('version')


class Cppcheck(analyzer_base.SourceAnalyzer):
    """
    Constructs the Cppcheck analyzer commands.
    """

    ANALYZER_NAME = 'cppcheck'

    @classmethod
    def analyzer_binary(cls):
        return analyzer_context.get_context() \
            .analyzer_binaries[cls.ANALYZER_NAME]

    @classmethod
    def get_binary_version(self, environ, details=False) -> str:
        """ Get analyzer version information. """
        # No need to LOG here, we will emit a warning later anyway.
        if not self.analyzer_binary():
            return None
        version = [self.analyzer_binary(), '--version']
        try:
            output = subprocess.check_output(version,
                                             env=environ,
                                             universal_newlines=True,
                                             encoding="utf-8",
                                             errors="ignore")
            if details:
                return output.strip()
            return parse_version(output)
        except (subprocess.CalledProcessError, OSError) as oerr:
            LOG.warning("Failed to get analyzer version: %s",
                        ' '.join(version))
            LOG.warning(oerr)

        return None

    def add_checker_config(self, checker_cfg):
        LOG.error("Checker configuration for Cppcheck is not implemented yet")

    def get_analyzer_mentioned_files(self, output):
        """
        Return a collection of files that were mentioned by the analyzer in
        its standard outputs, which should be analyzer_stdout or
        analyzer_stderr from a result handler.
        """
        pass

    def parse_analyzer_config(self):
        """
        Parses a set of a white listed compiler flags.
        Cppcheck can only use a subset of the parametes
        found in compilation commands.
        These are:
        * -I: flag for specifing include directories
        * -D: for build time defines
        * -U: for undefining names
        * -std: The languange standard
        Any other parameter different from the above list will be dropped.
        """
        params = []
        interesting_option = re.compile("-[IUD].*")
        # the std flag is different. the following are all valid flags:
        # * --std c99
        # * -std=c99
        # * --std=c99
        # BUT NOT:
        # * -std c99
        # * -stdlib=libc++
        std_regex = re.compile("-(-std$|-?std=.*)")
        for i, analyzer_option in enumerate(self.buildaction.analyzer_options):
            if interesting_option.match(analyzer_option):
                params.extend([analyzer_option])
                # The above extend() won't properly insert the analyzer_option
                # in case of the following format -I <path/to/include>.
                # The below check will add the next item in the
                # analyzer_options list if the parameter is specified with a
                # space, as that should be actual path to the include.
                if interesting_option.match(analyzer_option).span() == (0, 2):
                    params.extend(
                        [self.buildaction.analyzer_options[i+1]]
                    )
            elif std_regex.match(analyzer_option):
                standard = ""
                if "=" in analyzer_option:
                    standard = analyzer_option.split("=")[-1]
                # Handle space separated parameter
                # The else clause is never executed until a log parser
                # limitation is addressed, as only this "-std=xxx" version
                # of the paramter is forwareded in the analyzer_option list.
                else:
                    standard = self.buildaction.analyzer_options[i+1]
                standard = standard.lower().replace("gnu", "c")
                params.extend(["--std=" + standard])
        return params

    def construct_analyzer_cmd(self, result_handler):
        """
        Construct analyzer command for cppcheck.
        """
        try:
            config = self.config_handler

            analyzer_cmd = [Cppcheck.analyzer_binary()]

            # TODO implement a more granular commandline checker config
            # Cppcheck runs with all checkers enabled for the time being
            # the unneded checkers are passed as suppressed checkers
            analyzer_cmd.append('--enable=all')

            for checker_name, value in config.checks().items():
                if value[0] == CheckerState.disabled:
                    # TODO python3.9 removeprefix method would be nicer
                    # than startswith and a hardcoded slicing
                    if checker_name.startswith("cppcheck-"):
                        checker_name = checker_name[9:]
                    analyzer_cmd.append('--suppress=' + checker_name)

            # unusedFunction check is for whole program analysis,
            # which is not compatible with per source file analysis.
            analyzer_cmd.append('--suppress=unusedFunction')

            # Add extra arguments.
            analyzer_cmd.extend(config.analyzer_extra_arguments)

            # Pass whitelisted parameters
            analyzer_cmd.extend(self.parse_analyzer_config())

            # TODO fix this in a follow up patch, because it is failing
            # the macos pypy test.
            # analyzer_cmd.extend(["--std=" +
            #                    self.buildaction.compiler_standard
            #                    .split("=")[-1].lower().replace("gnu", "c")])

            # By default there is no platform configuration,
            # but a platform definition xml can be specified.
            if 'platform' in config.analyzer_config:
                analyzer_cmd.append(
                        "--platform=" + config.analyzer_config['platform'][0])
            else:
                analyzer_cmd.append("--platform=native")

            if 'inconclusive' in config.analyzer_config:
                analyzer_cmd.append("--inconclusive")

            if 'addons' in config.analyzer_config:
                analyzer_cmd.extend(
                    f"--addon={addon}" for addon
                    in config.analyzer_config["addons"])

            if 'libraries' in config.analyzer_config:
                analyzer_cmd.extend(
                    f"--library={library}" for library
                    in config.analyzer_config["libraries"])

            # TODO Suggest a better place for this
            # cppcheck wont create the output folders for itself
            output_dir = Path(result_handler.workspace, "cppcheck",
                              result_handler.buildaction_hash)
            output_dir.mkdir(exist_ok=True)

            analyzer_cmd.append('--plist-output=' + str(output_dir))

            analyzer_cmd.append(self.source_file)

            return analyzer_cmd

        except Exception as ex:
            LOG.error(ex)
            return []

    @classmethod
    def get_analyzer_checkers(cls):
        """
        Return the list of the supported checkers.
        """
        command = [cls.analyzer_binary(), "--errorlist"]

        try:
            result = subprocess.check_output(command)
            return parse_checkers(result)
        except (subprocess.CalledProcessError) as e:
            LOG.error(e.stderr)
        except (OSError) as e:
            LOG.error(e.errno)
        return []

    @classmethod
    def get_analyzer_config(cls):
        """
        Config options for cppcheck.
        """
        return [("addons", "A list of cppcheck addon files."),
                ("libraries", "A list of cppcheck library definiton files."),
                ("platform", "The platform configuration .xml file."),
                ("inconclusive", "Enable inconclusive reports.")
                ]

    @classmethod
    def get_checker_config(cls):
        """
        TODO add config options for cppcheck checkers.
        """
        return []

    def analyze(self, analyzer_cmd, res_handler, proc_callback=None):
        env = None

        original_env_file = os.environ.get(
            'CODECHECKER_ORIGINAL_BUILD_ENV')
        if original_env_file:
            with open(original_env_file, 'rb') as env_file:
                env = pickle.load(env_file, encoding='utf-8')

        return super().analyze(analyzer_cmd, res_handler, proc_callback, env)

    def post_analyze(self, result_handler):
        """
        Post process the reuslts after the analysis.
        Will copy the plist files created by cppcheck into the
        root of the reports folder.
        Renames the source plist files to *.plist.bak because
        The report parsing of the Parse command is done recursively.

        """
        # Cppcheck can generate an id into the output plist file name
        # we get "the" *.plist file from the unique output folder

        cppcheck_out_folder = Path(result_handler.workspace,
                                   "cppcheck",
                                   result_handler.buildaction_hash)
        cppcheck_outs = cppcheck_out_folder.glob('**/*.plist')

        for cppcheck_out in list(cppcheck_outs):
            codechecker_out = os.path.join(result_handler.workspace,
                                           result_handler.analyzer_result_file)

            # plists generated by cppcheck are still "parsable" by our plist
            # parser. Renaming is needed to circumvent the processing of the
            # raw files.
            try:
                shutil.copy2(cppcheck_out, codechecker_out)
                Path(cppcheck_out).rename(str(cppcheck_out) + ".bak")
            except(OSError) as e:
                LOG.error(e.errno)

    @classmethod
    def resolve_missing_binary(cls, configured_binary, env):
        """
        In case of the configured binary for the analyzer is not found in the
        PATH, this method is used to find a callable binary.
        """

        LOG.debug("%s not found in path for Cppcheck!", configured_binary)

        if os.path.isabs(configured_binary):
            # Do not autoresolve if the path is an absolute path as there
            # is nothing we could auto-resolve that way.
            return False

        cppcheck = get_binary_in_path(['cppcheck'],
                                      r'^cppcheck(-\d+(\.\d+){0,2})?$',
                                      env)

        if cppcheck:
            LOG.debug("Using '%s' for Cppcheck!", cppcheck)
        return cppcheck

    @classmethod
    def is_binary_version_incompatible(cls, environ):
        """
        Check the version compatibility of the given analyzer binary.
        """
        analyzer_version = cls.get_binary_version(environ)

        # The analyzer version should be above 1.80 because '--plist-output'
        # argument was introduced in this release.
        if StrictVersion(analyzer_version) >= StrictVersion("1.80"):
            return None

        return "CppCheck binary found is too old at " \
               f"v{str(analyzer_version).strip()}; minimum version is 1.80"

    def construct_result_handler(self, buildaction, report_output,
                                 skiplist_handler):
        """
        See base class for docs.
        """
        res_handler = CppcheckResultHandler(buildaction, report_output,
                                            self.config_handler.report_hash)

        res_handler.skiplist_handler = skiplist_handler

        return res_handler

    @classmethod
    def construct_config_handler(cls, args):
        context = analyzer_context.get_context()
        handler = CppcheckConfigHandler()

        analyzer_config = defaultdict(list)

        if 'analyzer_config' in args and \
                isinstance(args.analyzer_config, list):
            for cfg in args.analyzer_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    analyzer_config[cfg.option].append(cfg.value)

        handler.analyzer_config = analyzer_config

        try:
            with open(args.cppcheck_args_cfg_file, 'r', encoding='utf8',
                      errors='ignore') as sa_cfg:
                handler.analyzer_extra_arguments = \
                    re.sub(r'\$\((.*?)\)',
                           env.replace_env_var(args.cppcheck_args_cfg_file),
                           sa_cfg.read().strip())
                handler.analyzer_extra_arguments = \
                    shlex.split(handler.analyzer_extra_arguments)
        except IOError as ioerr:
            LOG.debug_analyzer(ioerr)
        except AttributeError as aerr:
            # No cppcheck arguments file was given in the command line.
            LOG.debug_analyzer(aerr)

        check_env = context.analyzer_env

        # Overwrite PATH to contain only the parent of the cppcheck binary.
        if os.path.isabs(Cppcheck.analyzer_binary()):
            check_env['PATH'] = os.path.dirname(Cppcheck.analyzer_binary())

        checkers = cls.get_analyzer_checkers()

        # Cppcheck can and will report with checks that have a different
        # name than marked in the --errorlist xml. To be able to suppress
        # these reports, the checkerlist needs to be extended by those found
        # in the label file.
        checker_labels = context.checker_labels
        checkers_from_label = checker_labels.checkers("cppcheck")
        parsed_set = set([data[0] for data in checkers])
        for checker in set(checkers_from_label):
            if checker not in parsed_set:
                checkers.append((checker, ""))

        try:
            cmdline_checkers = args.ordered_checkers
        except AttributeError:
            LOG.debug_analyzer('No checkers were defined in '
                               'the command line for %s',
                               cls.ANALYZER_NAME)
            cmdline_checkers = []

        handler.initialize_checkers(
            checkers,
            cmdline_checkers,
            'enable_all' in args and args.enable_all)

        return handler



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Run pre analysis, collect statistics or CTU data.
"""

import os
import shlex
import shutil
import signal
import sys
import traceback
import uuid

import multiprocess

from codechecker_common.logger import get_logger

from codechecker_statistics_collector import post_process_stats

from .analyzers import analyzer_base
from .analyzers.clangsa import ctu_manager, ctu_triple_arch
from .analyzers.clangsa import statistics
from .analyzers.clangsa.analyzer import ClangSA

LOG = get_logger('analyzer')


def collect_statistics(action, source, clangsa_config, statistics_data):
    """
    Run the statistics collection command and save the
    stdout and stderr to a file.
    """
    cmd, can_collect = statistics.build_stat_coll_cmd(action, clangsa_config,
                                                      source)

    if not can_collect:
        LOG.debug('Can not collect statistical data.')
        return

    # TODO: shlex.join() will be more convenient in Python 3.8.
    LOG.debug_analyzer(' '.join(map(shlex.quote, cmd)))

    ret_code, analyzer_out, analyzer_err = \
        analyzer_base.SourceAnalyzer.run_proc(cmd)

    LOG.debug(analyzer_out)
    LOG.debug(analyzer_err)
    if ret_code:
        LOG.error("Failed to collect statistics for %s", source)
        return ret_code

    LOG.debug("Running statistics collectors for %s was sucesssful.",
              source)

    _, source_filename = os.path.split(source)

    output_id = source_filename + str(uuid.uuid4()) + '.stat'

    stat_for_source = os.path.join(statistics_data['stat_tmp_dir'],
                                   output_id)

    with open(stat_for_source, 'w', encoding="utf-8", errors="ignore") as out:
        out.write(analyzer_out)
        out.write(analyzer_err)

    return ret_code


# Progress reporting.
progress_checked_num = None
progress_actions = None


def init_worker(checked_num, action_num):
    global progress_checked_num, progress_actions
    progress_checked_num = checked_num
    progress_actions = action_num


def pre_analyze(params):
    action, clangsa_config, skip_handlers, ctu_data, statistics_data = params

    progress_checked_num.value += 1

    if skip_handlers and skip_handlers.should_skip(action.source):
        return
    if action.analyzer_type != ClangSA.ANALYZER_NAME:
        return

    _, source_filename = os.path.split(action.source)

    LOG.info("[%d/%d] %s",
             progress_checked_num.value,
             progress_actions.value, source_filename)

    try:
        if ctu_data:
            LOG.debug("running CTU pre analysis")
            ctu_temp_fnmap_folder = ctu_data.get('ctu_temp_fnmap_folder')
            ctu_func_map_cmd = ctu_data.get('ctu_func_map_cmd')

            triple_arch = \
                ctu_triple_arch.get_triple_arch(action, action.source,
                                                clangsa_config)

            # TODO: reorganize the various ctu modes parameters
            # Dump-based analysis requires serialized ASTs.
            if clangsa_config.ctu_on_demand:
                ctu_manager.generate_invocation_list(triple_arch, action,
                                                     action.source,
                                                     clangsa_config)
            else:
                ctu_manager.generate_ast(triple_arch, action, action.source,
                                         clangsa_config)
            # On-demand analysis does not require AST-dumps.
            # We map the function names to corresponding sources of ASTs.
            # In case of On-demand analysis this source is the original source
            # code. In case of AST-dump based analysis these sources are the
            # generated AST-dumps.
            ctu_manager.map_functions(triple_arch, action, action.source,
                                      clangsa_config, ctu_func_map_cmd,
                                      ctu_temp_fnmap_folder)

    except Exception as ex:
        LOG.error("Pre-analysis failed for %s: %s", action.source, str(ex))
        traceback.print_exc(file=sys.stdout)

    try:
        if statistics_data:
            LOG.debug("running statistics pre analysis")
            collect_statistics(action,
                               action.source,
                               clangsa_config,
                               statistics_data)

    except Exception as ex:
        LOG.debug_analyzer(str(ex))
        traceback.print_exc(file=sys.stdout)
        raise


def run_pre_analysis(actions, clangsa_config,
                     jobs, skip_handlers, ctu_data, statistics_data, manager):
    """
    Run multiple pre analysis jobs before the actual analysis.
    """
    # pylint: disable=no-member multiprocess module members.
    LOG.info('Pre-analysis started.')
    if ctu_data:
        LOG.info("Collecting data for ctu analysis.")
    if statistics_data:
        LOG.info("Collecting data for statistical analysis.")

    def signal_handler(signum, frame):
        try:
            pool.terminate()
            manager.shutdown()
        finally:
            sys.exit(128 + signum)

    signal.signal(signal.SIGINT, signal_handler)

    processed_var = multiprocess.Value('i', 0)
    actions_num = multiprocess.Value('i', len(actions))

    pool = multiprocess.Pool(jobs,
                             initializer=init_worker,
                             initargs=(processed_var, actions_num))

    if statistics_data:
        # Statistics collection is enabled setup temporary
        # directories.
        stat_tmp_dir = statistics_data['stat_tmp_dir']

        # Cleaning previous outputs.
        if os.path.exists(stat_tmp_dir):
            shutil.rmtree(stat_tmp_dir)

        os.makedirs(stat_tmp_dir)

    try:
        collect_actions = [(build_action,
                            clangsa_config,
                            skip_handlers,
                            ctu_data,
                            statistics_data)
                           for build_action in actions]
        # FIXME: Ensure all shared data structures are wrapped in manager
        #        proxy objects before passing them to other processes via
        #        map_async.
        #        Note that even deep-copying is known to be insufficient.
        result = pool.map_async(pre_analyze, collect_actions)
        pool.close()
    except Exception:
        pool.terminate()
        raise
    finally:
        pool.join()
        # Return whether the call completed without raising an exception.
        if not result.successful():
            # If the remote call raised an exception then that exception will
            # be reraised by get().
            result.get()

    # Postprocessing the pre analysis results.
    if ctu_data:
        ctu_manager.merge_clang_extdef_mappings(
                ctu_data.get('ctu_dir'),
                ctu_data.get('ctu_func_map_file'),
                ctu_data.get('ctu_temp_fnmap_folder'))

    if statistics_data:

        stats_in = statistics_data.get('stat_tmp_dir')
        stats_out = statistics_data.get('stats_out_dir')

        post_process_stats.process(stats_in, stats_out,
                                   statistics_data.get(
                                      'stats_min_sample_count'),
                                   statistics_data.get(
                                      'stats_relevance_threshold'))

        if os.path.exists(stats_in):
            LOG.debug('Cleaning up temporary statistics directory')
            shutil.rmtree(stats_in)
    LOG.info('Pre-analysis finished.')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
This module stores constants that are shared between the CodeChecker server
and client, related to API and other version-specific information.
"""


# The name of the cookie which contains the user's authentication session's
# token.
SESSION_COOKIE_NAME = '__ccPrivilegedAccessToken'

# The newest supported minor version (value) for each supported major version
# (key) in this particular build.
SUPPORTED_VERSIONS = {
    6: 58
}

# Used by the client to automatically identify the latest major and minor
# version.
CLIENT_API = '{0}.{1}'.format(
    max(SUPPORTED_VERSIONS.keys()),
    SUPPORTED_VERSIONS[max(SUPPORTED_VERSIONS.keys())])


def get_version_str():
    return ', '.join(["v" + str(v) + "." + str(SUPPORTED_VERSIONS[v])
                      for v in SUPPORTED_VERSIONS])



"""
Authentication session tokens

Revision ID: 4964142b58d2
Revises:     6b9f832d0b20
Create Date: 2018-03-28 10:21:38.593302
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '4964142b58d2'
down_revision = '6b9f832d0b20'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'auth_sessions',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_name', sa.String(), nullable=True),
        sa.Column('token', sa.CHAR(length=32), nullable=False),
        sa.Column('groups', sa.String(), nullable=True),
        sa.Column('last_access', sa.DateTime(), nullable=False),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('can_expire',
                  sa.Boolean(),
                  server_default=sa.text('true'),
                  nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_auth_sessions')),
        sa.UniqueConstraint('token', name=op.f('uq_auth_sessions_token')))

    op.drop_table('user_sessions')


def downgrade():
    op.create_table(
        'user_sessions',
        sa.Column('user_name', sa.VARCHAR(), nullable=False),
        sa.Column('token', sa.CHAR(length=32), nullable=False),
        sa.Column('groups', sa.VARCHAR(), nullable=True),
        sa.Column('last_access', sa.DATETIME(), nullable=False),
        sa.PrimaryKeyConstraint('user_name', name='pk_user_sessions'))

    op.drop_table('auth_sessions')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Clang Static Analyzer related functions.
"""

import os
import plistlib
import re
import shlex
import subprocess

from typing import List

from codechecker_common.logger import get_logger

from codechecker_analyzer import analyzer_context, env
from codechecker_statistics_collector.collectors.special_return_value import \
    SpecialReturnValueCollector
from codechecker_statistics_collector.collectors.return_value import \
    ReturnValueCollector

from .. import analyzer_base
from ..config_handler import CheckerState
from ..flag import has_flag
from ..flag import prepend_all

from . import config_handler
from . import ctu_triple_arch
from . import version
from .ctu_autodetection import CTUAutodetection
from .result_handler import ClangSAResultHandler

LOG = get_logger('analyzer')


def parse_clang_help_page(
    command: List[str],
    start_label: str
) -> List[str]:
    """
    Parse the clang help page starting from a specific label.
    Returns a list of (flag, description) tuples.
    """
    try:
        help_page = subprocess.check_output(
            command,
            stderr=subprocess.STDOUT,
            env=analyzer_context.get_context().analyzer_env,
            universal_newlines=True,
            encoding="utf-8",
            errors="ignore")
    except (subprocess.CalledProcessError, OSError):
        LOG.debug("Failed to run '%s' command!", command)
        return []

    try:
        help_page = help_page[help_page.index(start_label) + len(start_label):]
    except ValueError:
        return []

    # This regex will match lines which contain only a flag or a flag and a
    # description: '  <flag>', '  <flag> <description>'.
    start_new_option_rgx = \
        re.compile(r"^\s{2}(?P<flag>\S+)(\s(?P<desc>[^\n]+))?$")

    # This regex will match lines which contain description for the previous
    # flag: '     <description>'
    prev_help_desc_rgx = \
        re.compile(r"^\s{3,}(?P<desc>[^\n]+)$")

    res = []

    flag = None
    desc = []
    for line in help_page.splitlines():
        m = start_new_option_rgx.match(line)
        if m:
            if flag and desc:
                res.append((flag, ' '.join(desc)))
                flag = None
                desc = []

            flag = m.group("flag")
        else:
            m = prev_help_desc_rgx.match(line)

        if m and m.group("desc"):
            desc.append(m.group("desc").strip())

    if flag and desc:
        res.append((flag, ' '.join(desc)))

    return res


def _is_user_disabled_checker(checker, ordered_checkers):
    """
    This function returns True if the given checker is disabled by the user
    explicitly by a --disable flag.
    """
    if not ordered_checkers:
        return False

    disabled_checkers = (c for c, enabled in ordered_checkers if not enabled)

    return any(checker.startswith(c) for c in disabled_checkers)


class ClangSA(analyzer_base.SourceAnalyzer):
    """
    Constructs clang static analyzer commands.
    """
    ANALYZER_NAME = 'clangsa'

    __ctu_autodetection = None

    def __init__(self, cfg_handler, buildaction):
        super(ClangSA, self).__init__(cfg_handler, buildaction)
        self.__disable_ctu = False
        self.__checker_configs = []
        self.__disabled_checkers = []

    @classmethod
    def analyzer_binary(cls):
        """
        Return the path of the analyzer binary.
        """
        return analyzer_context.get_context() \
            .analyzer_binaries[cls.ANALYZER_NAME]

    @classmethod
    def analyzer_plugins(cls) -> List[str]:
        """
        Return the list of .so file paths which contain checker plugins to
        ClangSA.
        """
        plugin_dir = analyzer_context.get_context().checker_plugin

        clangsa_plugin_dir = env.get_clangsa_plugin_dir()
        is_analyzer_from_path = env.is_analyzer_from_path()
        if is_analyzer_from_path:
            if not clangsa_plugin_dir:
                return []

            # If the CC_ANALYZERS_FROM_PATH and CC_CLANGSA_PLUGIN_DIR
            # environment variables are set we will use this value as the
            # plugin directory.
            plugin_dir = clangsa_plugin_dir

        if not plugin_dir or not os.path.exists(plugin_dir):
            return []

        return [os.path.join(plugin_dir, f)
                for f in os.listdir(plugin_dir)
                if os.path.isfile(os.path.join(plugin_dir, f))
                and f.endswith(".so")]

    @classmethod
    def __add_plugin_load_flags(cls, analyzer_cmd: List[str]):
        """
        ClangSA can be extended with checker plugins. This function extends a
        clang command with these plugins.
        """
        for plugin in ClangSA.analyzer_plugins():
            analyzer_cmd.extend(["-load", plugin])

    @classmethod
    def get_binary_version(self, environ, details=False) -> str:
        # No need to LOG here, we will emit a warning later anyway.
        if not self.analyzer_binary():
            return None

        if details:
            version = [self.analyzer_binary(), '--version']
        else:
            version = [self.analyzer_binary(), '-dumpversion']
        try:
            output = subprocess.check_output(version,
                                             env=environ,
                                             universal_newlines=True,
                                             encoding="utf-8",
                                             errors="ignore")
            return output.strip()
        except (subprocess.CalledProcessError, OSError) as oerr:
            LOG.warning("Failed to get analyzer version: %s",
                        ' '.join(version))
            LOG.warning(oerr)

        return None

    @classmethod
    def version_info(cls):
        return version.get(cls.analyzer_binary())

    @classmethod
    def ctu_capability(cls):
        """
        Return a CTUAutodetection object which describes the availability of
        CTU feature with some additional CTU-related info.
        """
        if not cls.__ctu_autodetection:
            cls.__ctu_autodetection = CTUAutodetection(
                cls.analyzer_binary(),
                analyzer_context.get_context().analyzer_env)

        return cls.__ctu_autodetection

    def is_ctu_available(self):
        """
        Check if ctu is available for the analyzer.
        If the ctu_dir is set in the config, the analyzer is capable to
        run ctu analysis.
        """
        return bool(self.config_handler.ctu_dir)

    def is_ctu_enabled(self):
        """
        Check if ctu is enabled for the analyzer.
        """
        return not self.__disable_ctu

    def disable_ctu(self):
        """
        Disable ctu even if ctu is available.
        By default it is enabled if available.
        """
        self.__disable_ctu = True

    def enable_ctu(self):
        self.__disable_ctu = False

    def add_checker_config(self, checker_cfg):
        """
        Add configuration options to specific checkers.
        checker_cfg should be a list of arguments in case of
        Clang Static Analyzer like this:
        ['-Xclang', '-analyzer-config', '-Xclang', 'checker_option=some_value']
        """

        self.__checker_configs.append(checker_cfg)

    @classmethod
    def get_analyzer_checkers(
        cls,
        alpha: bool = True,
        debug: bool = False
    ) -> List[str]:
        """
        Return the list of the supported checkers.

        Before clang9 alpha and debug checkers were printed by default. Since
        clang9 there are extra arguments to print the additional checkers.
        """
        command = [cls.analyzer_binary(), "-cc1"]

        cls.__add_plugin_load_flags(command)

        command.append("-analyzer-checker-help")

        # The clang compiler on OSX is a few
        # relases older than the open source clang release.
        # The new checker help printig flags are not available there yet.
        # If the OSX clang will be updated to based on clang v8
        # this early return can be removed.
        version_info = cls.version_info()
        if version_info and version_info.vendor == "clang":
            if alpha and version_info.major_version >= 9:
                command.append("-analyzer-checker-help-alpha")

            if debug and version_info.major_version >= 9:
                command.append("-analyzer-checker-help-developer")

        return parse_clang_help_page(command, 'CHECKERS:')

    @classmethod
    def get_checker_config(cls) -> List[str]:
        """
        Return the list of checker config options.

        Before clang9 alpha and debug checkers were printed by default. Since
        clang9 there are extra arguments to print the additional checkers.
        """
        command = [cls.analyzer_binary(), "-cc1"]

        cls.__add_plugin_load_flags(command)

        command.append("-analyzer-checker-option-help")

        version_info = ClangSA.version_info()
        if version_info.vendor == "clang":
            if version_info.major_version >= 9:
                command.append("-analyzer-checker-option-help-alpha")

            if version_info.major_version >= 9:
                command.append("-analyzer-checker-option-help-developer")

        return parse_clang_help_page(command, 'OPTIONS:')

    @classmethod
    def get_analyzer_config(cls) -> List[str]:
        """Return the list of analyzer config options."""
        command = [cls.analyzer_binary(), "-cc1"]

        cls.__add_plugin_load_flags(command)

        command.append("-analyzer-config-help")

        return parse_clang_help_page(command, 'OPTIONS:')

    def post_analyze(self, result_handler):
        """
        Disabled checkers are not actually disabled during analysis, because
        they might rely on each other under the hood. The disabled checkers'
        reports are removed in this post-processing step.
        """
        try:
            if not os.path.isfile(result_handler.analyzer_result_file):
                # This check has the same race-condition reason as the
                # exception, so see its description below.
                return

            with open(result_handler.analyzer_result_file, 'rb') as f:
                plist = plistlib.load(f)
        except plistlib.InvalidFileException:
            # It may happen that a compilation database contains a build action
            # multiple times (because it is compiled for several target
            # architectures), or at least they differ in a so minor part that
            # the same .plist file belongs to them. (For further details see
            # analyzer_action_str() in ResultHandler about the strange behavior
            # of make 4.3 in its -o flag.)
            # If this happens then the analysis of the same build action is
            # analyzed in parallel several times, so the same output .plist
            # file is changed by several threads. This may result an invalid
            # .plist which fails plistlib.load(). This is not a big problem,
            # because the second thread will also execute this post-processing
            # and it happens rarely anyways. test_compile_uniqueing() fails
            # undeterministically without this.
            return

        plist['diagnostics'] = list(filter(
            lambda diag: diag['check_name'] not in self.__disabled_checkers,
            plist.get('diagnostics', [])))

        with open(result_handler.analyzer_result_file, 'wb') as f:
            plistlib.dump(plist, f)

    def construct_analyzer_cmd(self, result_handler):
        """
        Called by the analyzer method.
        Construct the analyzer command.
        """
        try:
            # Get an output file from the result handler.
            analyzer_output_file = result_handler.analyzer_result_file

            # Get the checkers list from the config_handler.
            # Checker order matters.
            config = self.config_handler

            analyzer_cmd = [ClangSA.analyzer_binary(), '--analyze',
                            # Do not warn about the unused gcc/g++ arguments.
                            '-Qunused-arguments']

            for plugin in ClangSA.analyzer_plugins():
                analyzer_cmd.extend(["-Xclang", "-plugin",
                                     "-Xclang", "checkercfg",
                                     "-Xclang", "-load",
                                     "-Xclang", plugin])

            analyzer_mode = 'plist-multi-file'
            analyzer_cmd.extend(['-Xclang',
                                 '-analyzer-opt-analyze-headers',
                                 '-Xclang',
                                 '-analyzer-output=' + analyzer_mode,
                                 '-o', analyzer_output_file])

            # Expand macros in plist output on the bug path.
            analyzer_cmd.extend(['-Xclang',
                                 '-analyzer-config',
                                 '-Xclang',
                                 'expand-macros=true'])

            # Checker configuration arguments needs to be set before
            # the checkers.
            if self.__checker_configs:
                for cfg in self.__checker_configs:
                    analyzer_cmd.extend(cfg)

            # TODO: This object has a __checker_configs attribute and the
            # corresponding functions to set it. Either those should be used
            # for checker configs coming as command line argument, or those
            # should be eliminated.
            for cfg in config.checker_config:
                analyzer_cmd.extend(
                    ['-Xclang', '-analyzer-config', '-Xclang', cfg])

            # Config handler stores which checkers are enabled or disabled.
            self.__disabled_checkers = []
            enabled_checkers = []
            for checker_name, value in config.checks().items():
                state, _ = value
                if state == CheckerState.enabled:
                    enabled_checkers.append(checker_name)
                elif state == CheckerState.disabled:
                    self.__disabled_checkers.append(checker_name)

            if enabled_checkers:
                analyzer_cmd.extend(['-Xclang',
                                     '-analyzer-checker=' +
                                     ','.join(enabled_checkers)])
            # Enable aggressive-binary-operation-simplification option.
            version_info = ClangSA.version_info()
            if version_info and version_info.major_version >= 8:
                analyzer_cmd.extend([
                    '-Xclang',
                    '-analyzer-config',
                    '-Xclang',
                    'aggressive-binary-operation-simplification=true'])

            # Enable the z3 solver backend.
            if config.enable_z3:
                analyzer_cmd.extend(['-Xclang', '-analyzer-constraints=z3'])

            if config.enable_z3_refutation and not config.enable_z3:
                analyzer_cmd.extend(['-Xclang',
                                     '-analyzer-config',
                                     '-Xclang',
                                     'crosscheck-with-z3=true'])

            if config.ctu_dir and not self.__disable_ctu:
                analyzer_cmd.extend(
                    ['-Xclang', '-analyzer-config', '-Xclang',
                     'experimental-enable-naive-ctu-analysis=true',
                     '-Xclang', '-analyzer-config', '-Xclang',
                     'ctu-dir=' + self.get_ctu_dir()])
                ctu_display_progress = \
                    ClangSA.ctu_capability().display_progress
                if ctu_display_progress:
                    analyzer_cmd.extend(ctu_display_progress)

                if config.ctu_on_demand:
                    invocation_list_path = \
                        os.path.join(self.get_ctu_dir(), 'invocation-list.yml')
                    analyzer_cmd.extend(
                        ['-Xclang', '-analyzer-config', '-Xclang',
                         f'ctu-invocation-list={invocation_list_path}'
                         ])

            compile_lang = self.buildaction.lang
            if not has_flag('-x', analyzer_cmd):
                analyzer_cmd.extend(['-x', compile_lang])

            if not has_flag('--target', analyzer_cmd) and \
                    self.buildaction.target != "":
                analyzer_cmd.append(f"--target={self.buildaction.target}")

            if not has_flag('-arch', analyzer_cmd) and \
                    self.buildaction.arch != "":
                analyzer_cmd.extend(["-arch", self.buildaction.arch])

            if not has_flag('-std', analyzer_cmd) and \
                    self.buildaction.compiler_standard != "":
                analyzer_cmd.append(self.buildaction.compiler_standard)

            analyzer_cmd.extend(config.analyzer_extra_arguments)

            analyzer_cmd.extend(self.buildaction.analyzer_options)

            analyzer_cmd.extend(prepend_all(
                '-isystem',
                self.buildaction.compiler_includes))

            analyzer_cmd.append(self.source_file)

            return analyzer_cmd

        except Exception as ex:
            LOG.error(ex)
            return []

    def get_ctu_dir(self):
        """
        Returns the path of the ctu directory (containing the triple).
        """
        config = self.config_handler
        triple_arch = ctu_triple_arch.get_triple_arch(self.buildaction,
                                                      self.source_file,
                                                      config)
        ctu_dir = os.path.join(config.ctu_dir, triple_arch)
        return ctu_dir

    def analyzer_mentioned_file_real_path(self, mentioned_path):
        """
        PCH-based an On-demand-parsed CTU modes use different paths and file
        suffixes. PCH-based mode uses ast dump files that are suffixed with
        '.ast', and they are supposed to be under the
        '<ctu-dir>/ast/<original-full-path>'. On-demand-parsed mode uses the
        full paths of the original source files.
        """
        pch_suffix = '.ast'

        # We convert the given file path to absolute path because we suppose
        # that in the clang's output the PCH files in CTU mode are relative
        # paths.
        mentioned_path = os.path.join(self.get_ctu_dir(), mentioned_path)

        # Detect the mode based on the path.
        suffix_index = mentioned_path.rfind(pch_suffix)
        # If the file does not have the suffix, the mode is On-demand-parsed.
        # Return the original path.
        if suffix_index == -1:
            LOG.debug("Analyzer mentioned path path: '%s', "
                      "corresponding source file: '%s'",
                      mentioned_path, mentioned_path)
            return mentioned_path

        # PCH-based mode stores files with their full path structure recreated
        # under <ctu-dir>/ast.
        ctu_ast_dir = os.path.join(self.get_ctu_dir(), 'ast')

        source_path = mentioned_path[len(ctu_ast_dir):suffix_index]

        LOG.debug("Analyzer mentioned path path: '%s', "
                  "corresponding source file: '%s'",
                  mentioned_path, source_path)

        if not mentioned_path.startswith(ctu_ast_dir):
            LOG.error(
                "Mentioned path '%s' ends with suffix '%s', but does "
                "not begin with supposed ast dir '%s'.", mentioned_path,
                pch_suffix, ctu_ast_dir)

        # Strip the prefix ast directory and the suffix.
        return mentioned_path[len(ctu_ast_dir):suffix_index]

    def get_analyzer_mentioned_files(self, output):
        """
        Parse ClangSA's output to generate a list of files that were mentioned
        in the standard output or standard error.
        """
        if not output:
            return set()

        regex_for_ctu_ast_load = re.compile(
            r"CTU loaded AST file: (.*)")

        paths = set()

        for line in output.splitlines():
            match = re.match(regex_for_ctu_ast_load, line)
            if match:
                path = match.group(1)
                paths.add(self.analyzer_mentioned_file_real_path(path))

        return paths

    @classmethod
    def resolve_missing_binary(cls, configured_binary, environ):
        """
        In case of the configured binary for the analyzer is not found in the
        PATH, this method is used to find a callable binary.
        """

        LOG.debug("%s not found in path for ClangSA!", configured_binary)

        if os.path.isabs(configured_binary):
            # Do not autoresolve if the path is an absolute path as there
            # is nothing we could auto-resolve that way.
            return False

        # clang, clang-5.0, clang++, clang++-5.1, ...
        clang = env.get_binary_in_path(['clang', 'clang++'],
                                       r'^clang(\+\+)?(-\d+(\.\d+){0,2})?$',
                                       environ)

        if clang:
            LOG.debug("Using '%s' for ClangSA!", clang)
        return clang

    @classmethod
    def is_binary_version_incompatible(cls, environ):
        """
        We support pretty much every ClangSA version.
        """
        return None

    def construct_result_handler(self, buildaction, report_output,
                                 skiplist_handler):
        """
        See base class for docs.
        """
        res_handler = ClangSAResultHandler(buildaction, report_output,
                                           self.config_handler.report_hash)

        res_handler.skiplist_handler = skiplist_handler

        return res_handler

    @classmethod
    def construct_config_handler(cls, args):

        context = analyzer_context.get_context()
        environ = context.analyzer_env

        handler = config_handler.ClangSAConfigHandler(environ)

        handler.report_hash = args.report_hash \
            if 'report_hash' in args else None

        handler.enable_z3 = 'enable_z3' in args and args.enable_z3 == 'on'

        handler.enable_z3_refutation = 'enable_z3_refutation' in args and \
            args.enable_z3_refutation == 'on'

        if 'ctu_phases' in args:
            handler.ctu_dir = os.path.join(args.output_path,
                                           args.ctu_dir)
            handler.ctu_on_demand = \
                'ctu_ast_mode' in args and \
                args.ctu_ast_mode == 'parse-on-demand'

        try:
            with open(args.clangsa_args_cfg_file, 'r', encoding='utf8',
                      errors='ignore') as sa_cfg:
                handler.analyzer_extra_arguments = \
                    re.sub(r'\$\((.*?)\)',
                           env.replace_env_var(args.clangsa_args_cfg_file),
                           sa_cfg.read().strip())
                handler.analyzer_extra_arguments = \
                    shlex.split(handler.analyzer_extra_arguments)
        except IOError as ioerr:
            LOG.debug_analyzer(ioerr)
        except AttributeError as aerr:
            # No clangsa arguments file was given in the command line.
            LOG.debug_analyzer(aerr)

        checkers = ClangSA.get_analyzer_checkers()

        try:
            cmdline_checkers = args.ordered_checkers
        except AttributeError:
            LOG.debug_analyzer('No checkers were defined in '
                               'the command line for %s', cls.ANALYZER_NAME)
            cmdline_checkers = []

        handler.initialize_checkers(
            checkers,
            cmdline_checkers,
            'enable_all' in args and args.enable_all)

        # If --stats flag is provided then enable statistics-based checkers
        # unless explicitly disabled by the user.
        if 'stats_enabled' in args:
            # In some Clang versions, statisticsbased checkers are in the alpha
            # package (alpha.statisticsbased), and maybe not in others. Lets
            # figure out the actual checker name first.
            special_return_value = next((
                checker for checker, _ in checkers
                if SpecialReturnValueCollector.checker_analyze in checker),
                None)
            unchecked_return_value = next((
                checker for checker, _ in checkers
                if ReturnValueCollector.checker_analyze in checker),
                None)

            if special_return_value and not _is_user_disabled_checker(
                    special_return_value,
                    cmdline_checkers):
                handler.set_checker_enabled(special_return_value)

            if unchecked_return_value and not _is_user_disabled_checker(
                    unchecked_return_value,
                    cmdline_checkers):
                handler.set_checker_enabled(unchecked_return_value)

        # Statistics collector checkers must be explicitly disabled as they
        # trash the output. These checkers are executed in a preceding analysis
        # phase.
        handler.set_checker_enabled(
            SpecialReturnValueCollector.checker_collect, False)

        handler.set_checker_enabled(
            ReturnValueCollector.checker_collect, False)

        handler.checker_config = []

        # TODO: This extra "isinstance" check is needed for
        # CodeChecker checkers --checker-config. This command also runs
        # this function in order to construct a config handler.
        if 'checker_config' in args and isinstance(args.checker_config, list):
            for cfg in args.checker_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    handler.checker_config.append(
                        f"{cfg.checker}:{cfg.option}={cfg.value}")

        # TODO: This extra "isinstance" check is needed for
        # CodeChecker analyzers --analyzer-config. This command also runs
        # this function in order to construct a config handler.
        if 'analyzer_config' in args and \
                isinstance(args.analyzer_config, list):
            for cfg in args.analyzer_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    handler.checker_config.append(f"{cfg.option}={cfg.value}")

        return handler



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging

from typing import Callable, Iterable, List, Optional, Set

from codechecker_report_converter.report import Report, SkipListHandlers
from codechecker_report_converter.report.hash import get_report_path_hash

LOG = logging.getLogger('report-converter')


class GenericSuppressHandler:
    get_suppressed: Callable[[Report], bool]
    store_suppress_bug_id: Callable[[str, str, str, str], bool]


def get_mentioned_original_files(reports: List[Report]) -> Set[str]:
    """ Get all mentioned files from the given reports. """
    files = set()

    for report in reports:
        files.update(report.original_files)

    return files


def get_changed_files(reports: List[Report]):
    """ Get all changed files from the given reports. """
    changed_files = set()

    for report in reports:
        changed_files.update(report.changed_files)

    return changed_files


def dump_changed_files(changed_files: Set[str]):
    """ Dump changed files. """
    if not changed_files:
        return None

    file_paths = '\n'.join([' - ' + f for f in changed_files])
    LOG.warning("The following source file contents changed or missing since "
                "the latest analysis:\n%s\nPlease re-analyze your "
                "project to update the reports!", file_paths)


def skip(
    reports: List[Report],
    processed_path_hashes: Optional[Set[str]] = None,
    skip_handlers: Optional[SkipListHandlers] = None,
    suppr_handler: Optional[GenericSuppressHandler] = None,
    review_status_filter: Optional[Iterable[str]] = None
) -> List[Report]:
    """ Skip reports. """
    kept_reports = []
    for report in reports:
        if skip_handlers and report.skip(skip_handlers):
            LOG.debug("Skip report because file path (%s) is on the skip "
                      "list.", report.file.path)
            continue

        if suppr_handler and suppr_handler.get_suppressed(report):
            LOG.debug("Suppressed by suppress file: %s:%s [%s] %s",
                      report.file.original_path, report.line,
                      report.checker_name, report.report_hash)
            continue

        if report.review_status:
            if suppr_handler:
                if not report.report_hash:
                    LOG.warning("Can't store suppress information for report "
                                "because no report hash is set: %s", report)
                    continue

                if report.review_status.status != 'unreviewed':
                    suppr_handler.store_suppress_bug_id(
                        report.report_hash,
                        report.file.name,
                        report.review_status.message.decode(),
                        report.review_status.status)

            if review_status_filter and \
                    report.review_status.status not in review_status_filter:
                LOG.debug("Filtered out by --review-status filter option: "
                          "%s:%s [%s] %s [%s]",
                          report.file.original_path, report.line,
                          report.checker_name, report.report_hash,
                          report.review_status)
                continue
        else:
            if review_status_filter and \
                    'unreviewed' not in review_status_filter:
                LOG.debug("Filtered out by --review-status filter option: "
                          "%s:%s [%s] %s [unreviewed]",
                          report.file.original_path, report.line,
                          report.checker_name, report.report_hash)
                continue

        if processed_path_hashes is not None:
            report_path_hash = get_report_path_hash(report)
            if report_path_hash in processed_path_hashes:
                LOG.debug("Not showing report because it is a deduplication "
                          "of an already processed report!")
                LOG.debug("Path hash: %s", report_path_hash)
                LOG.debug(report)
                continue

            processed_path_hashes.add(report_path_hash)

        kept_reports.append(report)

    return kept_reports



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import codechecker_api_shared.ttypes

from thrift.transport import TTransport
all_structs = []


class Confidentiality(object):
    CONFIDENTIAL = 0
    INTERNAL = 1
    OPEN = 2

    _VALUES_TO_NAMES = {
        0: "CONFIDENTIAL",
        1: "INTERNAL",
        2: "OPEN",
    }

    _NAMES_TO_VALUES = {
        "CONFIDENTIAL": 0,
        "INTERNAL": 1,
        "OPEN": 2,
    }


class DatabaseConnection(object):
    """
    Attributes:
     - engine
     - host
     - port
     - username_b64
     - password_b64
     - database
    """


    def __init__(self, engine=None, host=None, port=None, username_b64=None, password_b64=None, database=None,):
        self.engine = engine
        self.host = host
        self.port = port
        self.username_b64 = username_b64
        self.password_b64 = password_b64
        self.database = database

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.engine = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.host = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.port = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.username_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.password_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.STRING:
                    self.database = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('DatabaseConnection')
        if self.engine is not None:
            oprot.writeFieldBegin('engine', TType.STRING, 1)
            oprot.writeString(self.engine.encode('utf-8') if sys.version_info[0] == 2 else self.engine)
            oprot.writeFieldEnd()
        if self.host is not None:
            oprot.writeFieldBegin('host', TType.STRING, 2)
            oprot.writeString(self.host.encode('utf-8') if sys.version_info[0] == 2 else self.host)
            oprot.writeFieldEnd()
        if self.port is not None:
            oprot.writeFieldBegin('port', TType.I32, 3)
            oprot.writeI32(self.port)
            oprot.writeFieldEnd()
        if self.username_b64 is not None:
            oprot.writeFieldBegin('username_b64', TType.STRING, 4)
            oprot.writeString(self.username_b64.encode('utf-8') if sys.version_info[0] == 2 else self.username_b64)
            oprot.writeFieldEnd()
        if self.password_b64 is not None:
            oprot.writeFieldBegin('password_b64', TType.STRING, 5)
            oprot.writeString(self.password_b64.encode('utf-8') if sys.version_info[0] == 2 else self.password_b64)
            oprot.writeFieldEnd()
        if self.database is not None:
            oprot.writeFieldBegin('database', TType.STRING, 6)
            oprot.writeString(self.database.encode('utf-8') if sys.version_info[0] == 2 else self.database)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ProductConfiguration(object):
    """
    Attributes:
     - id
     - endpoint
     - displayedName_b64
     - description_b64
     - connection
     - runLimit
     - isReviewStatusChangeDisabled
     - confidentiality
     - reportLimit
    """


    def __init__(self, id=None, endpoint=None, displayedName_b64=None, description_b64=None, connection=None, runLimit=None, isReviewStatusChangeDisabled=None, confidentiality=None, reportLimit=None,):
        self.id = id
        self.endpoint = endpoint
        self.displayedName_b64 = displayedName_b64
        self.description_b64 = description_b64
        self.connection = connection
        self.runLimit = runLimit
        self.isReviewStatusChangeDisabled = isReviewStatusChangeDisabled
        self.confidentiality = confidentiality
        self.reportLimit = reportLimit

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.id = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.endpoint = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.displayedName_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.description_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRUCT:
                    self.connection = DatabaseConnection()
                    self.connection.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I64:
                    self.runLimit = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.BOOL:
                    self.isReviewStatusChangeDisabled = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.I32:
                    self.confidentiality = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.I64:
                    self.reportLimit = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ProductConfiguration')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.I64, 1)
            oprot.writeI64(self.id)
            oprot.writeFieldEnd()
        if self.endpoint is not None:
            oprot.writeFieldBegin('endpoint', TType.STRING, 2)
            oprot.writeString(self.endpoint.encode('utf-8') if sys.version_info[0] == 2 else self.endpoint)
            oprot.writeFieldEnd()
        if self.displayedName_b64 is not None:
            oprot.writeFieldBegin('displayedName_b64', TType.STRING, 3)
            oprot.writeString(self.displayedName_b64.encode('utf-8') if sys.version_info[0] == 2 else self.displayedName_b64)
            oprot.writeFieldEnd()
        if self.description_b64 is not None:
            oprot.writeFieldBegin('description_b64', TType.STRING, 4)
            oprot.writeString(self.description_b64.encode('utf-8') if sys.version_info[0] == 2 else self.description_b64)
            oprot.writeFieldEnd()
        if self.connection is not None:
            oprot.writeFieldBegin('connection', TType.STRUCT, 5)
            self.connection.write(oprot)
            oprot.writeFieldEnd()
        if self.runLimit is not None:
            oprot.writeFieldBegin('runLimit', TType.I64, 6)
            oprot.writeI64(self.runLimit)
            oprot.writeFieldEnd()
        if self.isReviewStatusChangeDisabled is not None:
            oprot.writeFieldBegin('isReviewStatusChangeDisabled', TType.BOOL, 7)
            oprot.writeBool(self.isReviewStatusChangeDisabled)
            oprot.writeFieldEnd()
        if self.confidentiality is not None:
            oprot.writeFieldBegin('confidentiality', TType.I32, 8)
            oprot.writeI32(self.confidentiality)
            oprot.writeFieldEnd()
        if self.reportLimit is not None:
            oprot.writeFieldBegin('reportLimit', TType.I64, 9)
            oprot.writeI64(self.reportLimit)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Product(object):
    """
    Attributes:
     - id
     - endpoint
     - displayedName_b64
     - description_b64
     - connected
     - accessible
     - administrating
     - databaseStatus
     - runCount
     - latestStoreToProduct
     - runLimit
     - admins
     - runStoreInProgress
     - confidentiality
     - reportLimit
    """


    def __init__(self, id=None, endpoint=None, displayedName_b64=None, description_b64=None, connected=None, accessible=None, administrating=None, databaseStatus=None, runCount=None, latestStoreToProduct=None, runLimit=None, admins=None, runStoreInProgress=None, confidentiality=None, reportLimit=None,):
        self.id = id
        self.endpoint = endpoint
        self.displayedName_b64 = displayedName_b64
        self.description_b64 = description_b64
        self.connected = connected
        self.accessible = accessible
        self.administrating = administrating
        self.databaseStatus = databaseStatus
        self.runCount = runCount
        self.latestStoreToProduct = latestStoreToProduct
        self.runLimit = runLimit
        self.admins = admins
        self.runStoreInProgress = runStoreInProgress
        self.confidentiality = confidentiality
        self.reportLimit = reportLimit

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.id = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.endpoint = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.displayedName_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.description_b64 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.BOOL:
                    self.connected = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.BOOL:
                    self.accessible = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 7:
                if ftype == TType.BOOL:
                    self.administrating = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 8:
                if ftype == TType.I32:
                    self.databaseStatus = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 9:
                if ftype == TType.I64:
                    self.runCount = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 10:
                if ftype == TType.STRING:
                    self.latestStoreToProduct = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 11:
                if ftype == TType.I64:
                    self.runLimit = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 12:
                if ftype == TType.LIST:
                    self.admins = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.admins.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 13:
                if ftype == TType.LIST:
                    self.runStoreInProgress = []
                    (_etype9, _size6) = iprot.readListBegin()
                    for _i10 in range(_size6):
                        _elem11 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.runStoreInProgress.append(_elem11)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 14:
                if ftype == TType.I32:
                    self.confidentiality = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 15:
                if ftype == TType.I64:
                    self.reportLimit = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Product')
        if self.id is not None:
            oprot.writeFieldBegin('id', TType.I64, 1)
            oprot.writeI64(self.id)
            oprot.writeFieldEnd()
        if self.endpoint is not None:
            oprot.writeFieldBegin('endpoint', TType.STRING, 2)
            oprot.writeString(self.endpoint.encode('utf-8') if sys.version_info[0] == 2 else self.endpoint)
            oprot.writeFieldEnd()
        if self.displayedName_b64 is not None:
            oprot.writeFieldBegin('displayedName_b64', TType.STRING, 3)
            oprot.writeString(self.displayedName_b64.encode('utf-8') if sys.version_info[0] == 2 else self.displayedName_b64)
            oprot.writeFieldEnd()
        if self.description_b64 is not None:
            oprot.writeFieldBegin('description_b64', TType.STRING, 4)
            oprot.writeString(self.description_b64.encode('utf-8') if sys.version_info[0] == 2 else self.description_b64)
            oprot.writeFieldEnd()
        if self.connected is not None:
            oprot.writeFieldBegin('connected', TType.BOOL, 5)
            oprot.writeBool(self.connected)
            oprot.writeFieldEnd()
        if self.accessible is not None:
            oprot.writeFieldBegin('accessible', TType.BOOL, 6)
            oprot.writeBool(self.accessible)
            oprot.writeFieldEnd()
        if self.administrating is not None:
            oprot.writeFieldBegin('administrating', TType.BOOL, 7)
            oprot.writeBool(self.administrating)
            oprot.writeFieldEnd()
        if self.databaseStatus is not None:
            oprot.writeFieldBegin('databaseStatus', TType.I32, 8)
            oprot.writeI32(self.databaseStatus)
            oprot.writeFieldEnd()
        if self.runCount is not None:
            oprot.writeFieldBegin('runCount', TType.I64, 9)
            oprot.writeI64(self.runCount)
            oprot.writeFieldEnd()
        if self.latestStoreToProduct is not None:
            oprot.writeFieldBegin('latestStoreToProduct', TType.STRING, 10)
            oprot.writeString(self.latestStoreToProduct.encode('utf-8') if sys.version_info[0] == 2 else self.latestStoreToProduct)
            oprot.writeFieldEnd()
        if self.runLimit is not None:
            oprot.writeFieldBegin('runLimit', TType.I64, 11)
            oprot.writeI64(self.runLimit)
            oprot.writeFieldEnd()
        if self.admins is not None:
            oprot.writeFieldBegin('admins', TType.LIST, 12)
            oprot.writeListBegin(TType.STRING, len(self.admins))
            for iter12 in self.admins:
                oprot.writeString(iter12.encode('utf-8') if sys.version_info[0] == 2 else iter12)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.runStoreInProgress is not None:
            oprot.writeFieldBegin('runStoreInProgress', TType.LIST, 13)
            oprot.writeListBegin(TType.STRING, len(self.runStoreInProgress))
            for iter13 in self.runStoreInProgress:
                oprot.writeString(iter13.encode('utf-8') if sys.version_info[0] == 2 else iter13)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.confidentiality is not None:
            oprot.writeFieldBegin('confidentiality', TType.I32, 14)
            oprot.writeI32(self.confidentiality)
            oprot.writeFieldEnd()
        if self.reportLimit is not None:
            oprot.writeFieldBegin('reportLimit', TType.I64, 15)
            oprot.writeI64(self.reportLimit)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(DatabaseConnection)
DatabaseConnection.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'engine', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'host', 'UTF8', None, ),  # 2
    (3, TType.I32, 'port', None, None, ),  # 3
    (4, TType.STRING, 'username_b64', 'UTF8', None, ),  # 4
    (5, TType.STRING, 'password_b64', 'UTF8', None, ),  # 5
    (6, TType.STRING, 'database', 'UTF8', None, ),  # 6
)
all_structs.append(ProductConfiguration)
ProductConfiguration.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'id', None, None, ),  # 1
    (2, TType.STRING, 'endpoint', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'displayedName_b64', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'description_b64', 'UTF8', None, ),  # 4
    (5, TType.STRUCT, 'connection', [DatabaseConnection, None], None, ),  # 5
    (6, TType.I64, 'runLimit', None, None, ),  # 6
    (7, TType.BOOL, 'isReviewStatusChangeDisabled', None, None, ),  # 7
    (8, TType.I32, 'confidentiality', None, None, ),  # 8
    (9, TType.I64, 'reportLimit', None, None, ),  # 9
)
all_structs.append(Product)
Product.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'id', None, None, ),  # 1
    (2, TType.STRING, 'endpoint', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'displayedName_b64', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'description_b64', 'UTF8', None, ),  # 4
    (5, TType.BOOL, 'connected', None, None, ),  # 5
    (6, TType.BOOL, 'accessible', None, None, ),  # 6
    (7, TType.BOOL, 'administrating', None, None, ),  # 7
    (8, TType.I32, 'databaseStatus', None, None, ),  # 8
    (9, TType.I64, 'runCount', None, None, ),  # 9
    (10, TType.STRING, 'latestStoreToProduct', 'UTF8', None, ),  # 10
    (11, TType.I64, 'runLimit', None, None, ),  # 11
    (12, TType.LIST, 'admins', (TType.STRING, 'UTF8', False), None, ),  # 12
    (13, TType.LIST, 'runStoreInProgress', (TType.STRING, 'UTF8', False), None, ),  # 13
    (14, TType.I32, 'confidentiality', None, None, ),  # 14
    (15, TType.I64, 'reportLimit', None, None, ),  # 15
)
fix_spec(all_structs)
del all_structs



"""
Add columns for number of runs and latest run storage

Revision ID: 4db450cf38af
Revises:     7ceea9232211
Create Date: 2021-06-28 15:52:57.237509
"""
from logging import getLogger

from alembic import op
import sqlalchemy as sa

from codechecker_server.database import database
from codechecker_server.database.run_db_model import IDENTIFIER as RUN_META
from codechecker_web.shared import webserver_context


# Revision identifiers, used by Alembic.
revision = '4db450cf38af'
down_revision = '7ceea9232211'
branch_labels = None
depends_on = None


def upgrade():
    LOG = getLogger("migration/config")

    op.add_column(
        'products',
        sa.Column('latest_storage_date', sa.DateTime(), nullable=True))

    op.add_column(
        'products',
        sa.Column('num_of_runs', sa.Integer(),
                  server_default="0", nullable=True))

    try:
        product_con = op.get_bind()
        products = product_con.execute(
            "SELECT id, endpoint, connection FROM products").fetchall()

        context = webserver_context.get_context()
        for id_, endpoint, connection in products:
            sql_server = database.SQLServer.from_connection_string(
                connection, endpoint, RUN_META, context.run_migration_root)

            engine = sa.create_engine(sql_server.get_connection_string())
            conn = engine.connect()

            run_info = conn.execute("SELECT COUNT(*), MAX(date) FROM runs") \
                .fetchone()

            values = [f"num_of_runs={run_info[0]}"]
            if run_info[1]:
                values.append(f"latest_storage_date='{run_info[1]}'")

            product_con.execute(f"""
                UPDATE products
                SET {', '.join(values)}
                WHERE id={id_}
            """)
    except Exception as ex:
        LOG.error("Failed to fill product detail columns (num_of_runs, "
                  "latest_storage_date): %s", ex)


def downgrade():
    op.drop_column('products', 'num_of_runs')
    op.drop_column('products', 'latest_storage_date')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


import base64


def to_b64(string):
    """ encode a string to a base64 encoded string """
    return base64.b64encode(string.encode("utf-8")).decode(
        "utf-8", errors="ignore"
    )


def from_b64(string_b64):
    """ decode a b46 encoded string to a string """
    return base64.b64decode(string_b64.encode("utf-8")).decode(
        "utf-8", errors="ignore"
    )
